/***************************************************************************/
/*Name:          SCENT: Dictionary                                         */
/*Location:      Undefined                                                 */
/*Author:        Justin A. Strauss                                         */
/*Organization:  Kaiser Permanente Southern California                     */
/*Purpose:       Match concepts                                            */
/*Modified:                                                                */
/***************************************************************************/

%macro matchConcepts;

	%macro matchConcepts_Batch(data=,out=,start=,obs=);
	%if (%length(&data)>0) %then %let data=%sysfunc(dequote(&data));
	%if (%length(&out)>0) %then %let out=%sysfunc(dequote(&out));
	%if (%length(&start)>0) %then %let start=%sysfunc(dequote(&start));
	%if (%length(&obs)>0) %then %let obs=%sysfunc(dequote(&obs));

		%macro stackPush(stack="tokenStack",counter="_stackNum",debug="NO",text=);
		%if (%length(&stack)>0) %then %let stack=%sysfunc(dequote(&stack));
		%if (%length(&counter)>0) %then %let counter=%sysfunc(dequote(&counter));
		%if (%length(&debug)>0) %then %let debug=%sysfunc(dequote(&debug));
		%if (%length(&text)>0) %then %let text=%sysfunc(dequote(&text));
		&counter=&counter+1;
		%if (%upcase(&debug)~=NO) %then %do;
			putlog "[PUSH] " &text;
		%end;
		&stack..add();
		%mend stackPush;

		%macro stackPull(stack="tokenStack",counter="_stackNum",debug="NO",text=);
		%if (%length(&stack)>0) %then %let stack=%sysfunc(dequote(&stack));
		%if (%length(&counter)>0) %then %let counter=%sysfunc(dequote(&counter));
		%if (%length(&debug)>0) %then %let debug=%sysfunc(dequote(&debug));
		%if (%length(&text)>0) %then %let text=%sysfunc(dequote(&text));
		&stack..find();
		&stack..remove();
		&counter=&counter-1;
		%if (%upcase(&debug)~=NO) %then %do;
			putlog "[PULL] " &text;
		%end;
		%mend stackPull;

		%macro stackStatusMsg(counter="_tokenCount",debug="YES",text=);
		%if (%length(&counter)>0) %then %let counter=%sysfunc(dequote(&counter));
		%if (%length(&debug)>0) %then %let debug=%sysfunc(dequote(&debug));
		%if (%length(&text)>0) %then %let text=%sysfunc(dequote(&text));
		%if (%upcase(&debug)~=NO) %then %do;
			if (_tokenCount=1) then putlog "     " %quote(&text);
			else if (_tokenCount=2) then putlog "          " %quote(&text);
			else if (_tokenCount=3) then putlog "               " %quote(&text);
			else if (_tokenCount=4) then putlog "                    " %quote(&text);
		%end;
		%mend stackStatusMsg;

		%macro iterHasNext(iter=,key=,result=);
		%if (%length(&iter)>0) %then %let iter=%sysfunc(dequote(&iter));
		%if (%length(&key)>0) %then %let key=%sysfunc(dequote(&key));
		%if (%length(&result)>0) %then %let result=%sysfunc(dequote(&result));
		_tempKey=&key;
		&iter..setcur(key: _tempKey);
		&result=ifn(&iter..next(),0,1);
		&iter..setcur(key: _tempKey);
		%mend iterHasNext;

		%macro createConceptTokens();
		data conceptTokens(rename=(
					conceptCode=_conceptCode
					conceptTokenID=_conceptTokenID
					conceptTokenNum=_conceptTokenNum
					conceptTokenSign=_conceptTokenSign
					conceptDate=_conceptDate
					conceptSearchDate=_conceptSearchDate
					conceptToken=_conceptToken
				));
			set _dict.conceptTokens;
		run;
		proc sort data=conceptTokens nodupkey; by _conceptTokenID _conceptCode _conceptTokenNum _conceptTokenSign _conceptToken; run;
		%mend createConceptTokens;

	%createConceptTokens;

	data tempReports(
			compress=yes keep=reportKey lineNum conceptMatchDate _conceptCode conceptPermutation _matchTokenID _matchTokenNum token _conceptTokenSign
			rename=(_conceptCode=conceptCode _matchTokenID=conceptTokenID _conceptTokenSign=conceptTokenSign _matchTokenNum=tokenNum)
		);

		/*concept dictionary tokens, keyed by token id*/
		length _conceptTokenID _conceptTokenNum _conceptTokenSign _conceptDate _conceptSearchDate 8 _conceptCode $10 _conceptToken $25;
		call missing(_conceptTokenID, _conceptTokenNum, _conceptTokenSign, _conceptDate, _conceptSearchDate, _conceptCode, _conceptToken);
		declare hash conceptTokens(multidata:"yes", dataset:"conceptTokens");
		call conceptTokens.defineKey('_conceptTokenID');
		call conceptTokens.defineData('_conceptTokenID','_conceptCode','_conceptTokenNum','_conceptTokenSign','_conceptDate','_conceptSearchDate','_conceptToken');
		call conceptTokens.defineDone();

		/*concept dictionary tokens, keyed by concept code*/
		declare hash concepts(multidata:"yes", dataset:"conceptTokens");
		call concepts.defineKey('_conceptCode');
		call concepts.defineData('_conceptCode','_conceptTokenID','_conceptTokenNum','_conceptTokenSign','_conceptDate','_conceptSearchDate','_conceptToken');
		call concepts.defineDone();
		declare hiter conceptIter('concepts');

		/*intermediate hash for tokens of current text segment*/
		length _matchTokenID _matchTokenNum 8;
		declare hash matchTokens(multidata:"yes");
		call matchTokens.defineKey('_matchTokenID');
		call matchTokens.defineData('_matchTokenID','_matchTokenNum','token');
		call matchTokens.defineDone();
		call missing(_matchTokenID, _matchTokenNum);
		declare hiter matchTokenIter('matchTokens');

		/*intermediate hash for concept codes having one or more tokens from current text segment*/
		length _matchConceptCode $10;
		declare hash matchConceptCodes();
		call matchConceptCodes.defineKey('_matchConceptCode');
		call matchConceptCodes.defineData('_matchConceptCode');
		call matchConceptCodes.defineDone();
		call missing(_matchConceptCode);
		declare hiter matchConceptCodeIter('matchConceptCodes');

		/*hash table containing token matches, multiple token positions allowed*/
		declare hash conceptTokenMatches();
		declare hiter conceptTokenMatchIter;

		/*hash table containing hash tables of token matches with multiple token positions*/
		declare hash conceptTokenMatchHashes(ordered: 'a');
		call conceptTokenMatchHashes.defineKey('_matchTokenID');
		call conceptTokenMatchHashes.defineData('_matchTokenID','_conceptCode','conceptTokenMatches','conceptTokenMatchIter');
		call conceptTokenMatchHashes.defineDone();
		declare hiter conceptTokenMatchHashIter;

		/*hash table containing all token position permutations of matched concepts*/
		length conceptPermutation conceptMatchDate 8;
		format conceptMatchDate mmddyy10.;
		declare hash matchPermutations(ordered: 'a');
		call matchPermutations.defineKey('reportKey','lineNum','_conceptCode','conceptPermutation','_matchTokenID','_matchTokenNum');
		call matchPermutations.defineData('reportKey','lineNum','linePos','conceptMatchDate','_conceptCode','conceptPermutation','_matchTokenID','_matchTokenNum','token','_conceptTokenSign');
		call matchPermutations.defineDone();
		call missing (conceptMatchDate, conceptPermutation);
		declare hiter matchPermutationIter('matchPermutations');

		/*hash table containing token positions of current match position permutation*/
		declare hash tokenPath(ordered: 'a');
		call tokenPath.defineKey('_matchTokenID');
		call tokenPath.defineData('_matchTokenID','conceptPermutation','_matchTokenNum','token','_conceptTokenSign');
		call tokenPath.defineDone();
		declare hiter tokenPathIter('tokenPath');

		/*stack system for queued processing of concept token permutations*/
		length _stackNum _tokenCount 8;
		declare hash tokenStack();
		call tokenStack.defineKey('_stackNum');
		call tokenStack.defineData('_stackNum','_tokenCount','_matchTokenID','_matchTokenNum','conceptTokenMatches','conceptTokenMatchIter');
		call tokenStack.defineDone();
		call missing(_stackNum, _tokenCount);

		/*set current observation pointer*/
		length _obsOld _obsStart 8;
		_obsOld=&start;
		_obsStart=&start;

		/*>>> (reports) = report by report*/
		do until (eof);
			/*increase counter of processed tokens*/

			/*>>> (reports : lines) = line by line in current report*/
			do until (last.reportKey);

				/*clear intermediate hashes from previously processed line*/
				matchTokens.clear();
				matchConceptCodes.clear();

				/*>>> (reports : lines : tokens) = token by token in current line*/
				do until (last.lineNum);
					set &data(firstobs=&start) end=eof;
					by reportKey lineNum tokenNum conceptTokenID;

					_obsStart+1;

					/*add current token to hash*/
					matchTokens.add(key: conceptTokenID, data: conceptTokenID, data: tokenNum, data: token);

					/*add concepts having current token to hash*/
					conceptTokenHash=conceptTokens.find(key: conceptTokenID);
					do while(conceptTokenHash=0);
						matchConceptCodes.ref(key: _conceptCode, data: _conceptCode);
						conceptTokenHash=conceptTokens.find_next();
					end;

				end;
				/*<<< (reports : lines : tokens)*/

				/*>>> (reports : lines : concepts) = partially matched concepts*/
				do while (matchConceptCodeIter.next()=0);
					conceptTokenMatchHashes.clear();
					length matchConceptResult 8;

					/*>>> (reports : lines : concepts : tokens) = tokens of partially matched concepts*/
					conceptHash=concepts.find(key: _matchConceptCode);


					do while(conceptHash=0);
						matchTokenHash=matchTokens.find(key: _conceptTokenID);

						/*current concept token not found in text segment*/
						if (matchTokenHash~=0) then do;
							/*skip to next concept unless unmatched token was negating*/
							if (_conceptTokenSign>=0) then do;
								matchConceptResult=0;
								leave;
							end; else do;
								conceptHash=concepts.find_next();
								continue;
							end;
						end; else do;
							/*create child hash for matches of current concept token in text segment*/
							conceptTokenMatches=_new_ hash(multidata:"yes", ordered: 'a');
							conceptTokenMatchIter=_new_ hiter('conceptTokenMatches');
							conceptTokenMatches.definekey('_matchTokenNum');
							conceptTokenMatches.definedata('_matchTokenNum','_matchTokenID','_conceptTokenSign','token');
							conceptTokenMatches.definedone();
							/*find all matches of current concept token in text segment*/
							do while(matchTokenHash=0);
								matchConceptResult=1;
								conceptTokenMatches.add();
								matchTokenHash=matchTokens.find_next();
							end;
							/*add child hash to parent with hashes for all concept tokens*/
							conceptTokenMatchHashes.add();
							/*move to next concept token*/
							conceptHash=concepts.find_next();
						end;

					end;
					/*<<< (reports : lines : concepts : tokens)*/

					/*create iterator for parent hash of concept tokens*/
					conceptTokenMatchHashIter=_new_ hiter('conceptTokenMatchHashes');

					/*create permutations of token positions for fully matched concepts*/
					if (matchConceptResult) then do;
						
						/*initialize search variables*/
						_numTokens=conceptTokenMatchHashes.num_items;
						_tokenCount=1;
						_nextToken=0;
						_nextNum=0;
						_stackNum=0;
						conceptPermutation=0;
						conceptMatchDate=today();

						/*add first occurrence of first concept token to top of processing stack*/
						conceptTokenMatchHashIter.first();
						conceptTokenMatchIter.first();
						%stackPush;

						/*process until stack is empty and no additional occurrences of current token exist*/
						do while (tokenStack.num_items>0 | _nextToken);

							/*pull token from stack if no additional occurrences of current token*/
							if (tokenStack.num_items>0 & _nextToken=0) then do;
								%stackPull;

								/*check for remaining concept tokens and additional occurrences of pulled token*/
								%iterHasNext(iter="conceptTokenMatchHashIter",key="_matchTokenID",result="_nextToken");
								%iterHasNext(iter="conceptTokenMatchIter",key="_matchTokenNum",result="_nextNum");
							end;

							/*if no remaining concept tokens then....*/
							if (_tokenCount>=_numTokens) then do;
								_nextToken=0;
								_nextNum=0;

								/*process occurrences of final concept token*/
								_rc=conceptTokenMatchIter.first();
								do while (_rc=0);
									_rc=tokenPath.remove(key: _matchTokenID);
									tokenPath.add();

									/*update current permutation count for concept*/
									conceptPermutation=conceptPermutation+1;
									_nextPerm=conceptPermutation;

									/*output tokens and positions of current permutation*/
									do while (tokenPathIter.next()=0);
										conceptPermutation=_nextPerm;
										tokenPath.replace();
										matchPermutations.add();
									end;

									/*remove last token from permutation and process next occurrence*/
									tokenPath.remove();
									_rc=conceptTokenMatchIter.next();
								end;

							/*otherwise, if additional concept tokens remain then...*/
							end; else do;
								
								/*replace prior occurrence of current token*/
								_rc=tokenPath.remove(key: _matchTokenID);
								tokenPath.add();

								/*push current token occurence to top of stack*/
								if (_nextNum & conceptTokenMatchIter.next()=0) then do;
									%stackPush;
								end;

								/*move to next token in concept*/
								if (conceptTokenMatchHashIter.next()=0) then do;
									conceptTokenMatchIter.first();
									_tokenCount=_tokenCount+1;
									_nextToken=1;
									_nextNum=0;
								end;

							end;
						end;

						/*output all permutations of matched concept*/
						do while (matchPermutationIter.next()=0);
							output tempReports;
						end;

						/*clear utilized hash tables*/
						matchPermutations.clear();
						tokenStack.clear();
						tokenPath.clear();
					end;

					/*delete current concept hash/iterator children and parent hash iterator*/
					rc=conceptTokenMatchHashIter.first();
					do while (rc=0);
						conceptTokenMatchIter.delete();
						conceptTokenMatches.clear();
						conceptTokenMatches.delete();
						rc=conceptTokenMatchHashIter.next();
					end;
					conceptTokenMatchHashIter.delete();

				end;
				/*<<< (reports : lines : concepts)*/

			end;
			/*<<< (reports : lines)*/
			
			/*exit if: batch processing, completed current report, and over batch length*/
			%if (%length(&start)) %then %do;
				if (last.reportKey & (eof | ((_obsStart-_obsOld)>&obs))) then do;
					call symputx("obsStart",put(_obsStart,12.0),"G");
					stop;
				end;
			%end;

		end; stop;
		/*<<< (reports)*/

	run;

	proc sql;
		create table &out(compress=yes) as
			select
				reportKey
				,lineNum
				,conceptCode
				,(max(tokenNum)-min(tokenNum)-count(tokenNum)+1) as tokenDistance
				, min(tokenNum) as minToken, max(tokenNum) as maxToken
				,conceptPermutation
				,conceptTokenID
				,conceptTokenSign
				,tokenNum
				,token
				,conceptMatchDate
			from tempReports
			group by reportKey, lineNum, conceptCode, conceptPermutation
			order by reportKey, lineNum, conceptCode, tokenDistance, calculated minToken, calculated maxToken, conceptPermutation, tokenNum;
		alter table &out drop minToken, maxToken;
		drop table tempReports;
	quit;

	%mend matchConcepts_Batch;

/*determine number of reports*/
data _null_;
	if (0) then set &dataTokenMatches nobs=obsCount;
	call symputx("obsCount",put(obsCount,12.0),"G");
	call symputx("obsStart","1","G");
	stop;
run;

proc sort data=&dataTokenMatches out=tempTokens;
	by reportKey lineNum tokenNum conceptTokenID;
run;

/*batch process reports to counter slow memory leak of unknown origin*/
%let timeStart=%sysfunc(datetime());
%do %while (&obsStart<&obsCount);

	/*process ~1000 more tokens, stopping at first available report end*/
	%matchConcepts_Batch(data="tempTokens",start="&obsStart",obs="1000",out="tempPermutations");

	/*add new concept matches to dataset*/
	%if %sysfunc(exist(matchPermutations)) %then %do;
		data matchPermutations(compress=yes); set matchPermutations tempPermutations; run;
	%end; %else %do;
		data matchPermutations(compress=yes); set tempPermutations; run;
	%end;

	/*output log statistics*/
	%let timeNow=%sysfunc(datetime());
	%let timeLapse=%sysevalf(&timeNow-&timeStart, ceil);
	%let timeLapseHours=%sysfunc(putn(%sysevalf(&timeLapse/3600,floor),Z2.0));
	%let timeLapseMinutes=%sysfunc(putn(%sysevalf((&timeLapse/60)-(&timeLapseHours*60),floor),Z2.0));
	%let timeLapseSeconds=%sysfunc(putn(%sysevalf(&timeLapse-(&timeLapseHours*3600)-(&timeLapseMinutes*60),floor),Z2.0));
	%let pctDone=%sysevalf(100*((&obsStart-1)/&obscount),floor);
	%let timeRemain=%sysevalf(((&timeLapse/(&obsStart-1))*(&obsCount-(&obsStart-1))),floor);
	%let timeRemainHours=%sysfunc(putn(%sysevalf(&timeRemain/3600,floor),Z2.0));
	%let timeRemainMinutes=%sysfunc(putn(%sysevalf((&timeRemain/60)-(&timeRemainHours*60),floor),Z2.0));
	%let timeRemainSeconds=%sysfunc(putn(%sysevalf(&timeRemain-(&timeRemainHours*3600)-(&timeRemainMinutes*60),floor),Z2.0));
	%put %str( );
	%put NOTE: **********************************************************************;
	%put NOTE: [&timeLapseHours:&timeLapseMinutes:&timeLapseSeconds] (&pctDone.%) Processed %sysfunc(strip(%sysfunc(putn(&obsStart-1,comma12.0)))) of %sysfunc(strip(%sysfunc(putn(&obscount,comma12.0)))) report tokens;
	%put NOTE: Time remaining: &timeRemainHours hours, &timeRemainMinutes minutes, &timeRemainSeconds seconds.;
	%put NOTE: **********************************************************************;
	%put %str( );
	
%end;

/*create permanent dataset with concept matches*/
proc sort data=matchpermutations out=&dataConceptMatches(compress=yes) nodupkey;
	by reportKey lineNum conceptCode conceptPermutation conceptTokenID;
run;

/*update search date of clinical concepts*/
data _dict.conceptTokens(compress=yes);
	set _dict.conceptTokens;
	if (missing(conceptSearchDate)) then conceptSearchDate=datetime();
run;

/*clean up intermediate tables*/
proc sql;
	drop table tempPermutations, matchpermutations, tempTokens;
quit;

%mend matchConcepts;
