/***************************************************************************/
/*Name:          SCENT: Dictionary                                         */
/*Location:      Undefined                                                 */
/*Author:        Justin A. Strauss                                         */
/*Organization:  Kaiser Permanente Southern California                     */
/*Purpose:       Match negation patterns                                   */
/*Modified:                                                                */
/***************************************************************************/

/*NOTE: CURRENTLY INCOMPLETE*/

%macro matchNegation(data=, workbook=);
%if (%length(&data)>0) %then %let data=%sysfunc(dequote(&data));
%if (%length(&workbook)>0) %then %let workbook=%sysfunc(dequote(&workbook));

/*read in the negation patterns*/
libname xls "&workbook"  mixed=yes;
data negationPatterns;
	length _negationID _negationDate 8;
	format _negationDate DATETIME16.;
	set xls."SAS$"n(dbSasType=(pattern="char(500)" source="char(25)" class="char(25)")
		rename=(pattern=_negationText enabled=_negationEnabled source=_negationSource class=_negationClass prefix=_negationPrefix suffix=_negationSuffix));
	where (prxmatch("/\A(ALL|&sourceType)\z/i",strip(_negationSource)));
	length _negationPattern 8;
	_negationID=_N_;
	_negationDate=DateTime();
	call missing(_negationPattern);
	
run;
libname xls clear;

proc sort data=&dataTokenized(obs=10000) out=tempTokens(compress=yes);
	by reportKey lineNum tokenNum;
run;

data tempTokens(compress=yes);
	length _currentLine $2000;
	length _negationID _negationDate _negationPattern _negationPrefix _negationSuffix 8 _negationClass $25 _negationText $500;
	call missing(_negationID, _negationDate, _negationClass, _negationText, _negationPattern, _negationPrefix, _negationSuffix);
	declare hash negation(dataset:"negationPatterns");
	call negation.defineKey('_negationID');
	call negation.defineData('_negationID','_negationDate','_negationClass','_negationText','_negationPattern','_negationPrefix','_negationSuffix');
	call negation.defineDone();
	declare hiter negationIter('negation');
	negationHash=negationIter.first();
	do while (negationHash=0);
		_negationPattern=prxparse("/(\A|\b)" || strip(_negationText) || "(\z|\b)/");
		negation.replace();
		negationHash=negationIter.next();
	end;

	/*>>> (reports) = report by report*/
	do until (eof);
		/*increase counter of processed tokens*/

		/*>>> (reports : lines) = line by line in current report*/
		do until (last.reportKey);

			/*>>> (reports : lines : tokens) = token by token in current line*/
			_currentLine="";
			do until (last.lineNum);
				set tempTokens end=eof;
				by reportKey lineNum tokenNum;
				_currentLine=ifc(lengthn(_currentLine),catx(" ",_currentLine, token), strip(token));
			end;

			/**********************************************/
			/***** PLACE HOLDER FOR NEGATION MATCHING *****/
			/**********************************************/
			put " ";
			putlog _currentLine=;
			do while (negationIter.next()=0);
				if (prxmatch(_negationPattern,_currentLine)) then do;
					call prxposn(_negationPattern,0,_negationStart,_negationLen);
					_negationEnd=_negationStart+_negationLen;
					if (_negationStart=1) then negationStart=1;
					else do;
						negationStart=countw(substr(_currentLine,1,_negationStart-1))+1;
						negationStop=countw(substr(_currentLine,1,_negationStart-1))+1;
					end;
					
					putlog _negationStart= _negationEnd= _negationText=;
					putlog preText= negationStart=;
				end;
			end;


			/*<<< (reports : lines : tokens)*/

		end;
		/*<<< (reports : lines)*/

	end; stop;
	/*<<< (reports)*/

run;

%mend matchNegation;
