/***************************************************************************/
/*Name:          SCENT: Dictionary                                         */
/*Location:      Undefined                                                 */
/*Author:        Justin A. Strauss                                         */
/*Organization:  Kaiser Permanente Southern California                     */
/*Purpose:       Match concept tokens                                      */
/*Modified:                                                                */
/***************************************************************************/

%macro matchTokens;
data conceptTokenRegex(compress=yes);
	set _dict.conceptTokenRegex;
	length conceptTokenRegex 8;
	call missing(conceptTokenRegex);
run;
proc sql noprint;
	select max(conceptTokenDate) into :conceptTokenLatest from _dict.conceptTokenRegex;
quit;
data &dataTokenized(compress=yes keep=reportKey lineNum linePos tokenDate tokenIgnore tokenNum tokenPos tokenLen token)
		_tokenMatches(compress=yes keep=reportKey lineNum linePos tokenDate tokenIgnore tokenNum tokenPos tokenLen token conceptTokenID conceptToken);
	set &dataTokenized nobs=logCount;

	retain logLastPct logStart;
	if (_N_=1) then do;
		logStart=datetime();
		logLastPct=0;
		length conceptTokenID conceptTokenDate 8 conceptToken $200 conceptTokenRegex 8;
		declare hash conceptTokens(dataset:"conceptTokenRegex");
		call conceptTokens.defineKey('conceptTokenID');
		call conceptTokens.defineData('conceptTokenID');
		call conceptTokens.defineData('conceptTokenDate');
		call conceptTokens.defineData('conceptToken');
		call conceptTokens.defineData('conceptTokenRegex');
		call conceptTokens.defineDone();
		call missing(conceptTokenID, conceptTokenDate, conceptToken, conceptTokenRegex);
		declare hiter conceptTokenIter('conceptTokens');
		conceptTokenHash=conceptTokenIter.first();
		do while (conceptTokenHash=0);
			call conceptTokens.replace(key: conceptTokenID, data: conceptTokenID, data: conceptTokenDate, data: conceptToken, data: prxparse("/\A" || strip(conceptToken) || "\z/i"));
			conceptTokenHash=conceptTokenIter.next();
		end;
	end;
	if (missing(tokenDate)=0) then prevTokenDate=tokenDate;
	tokenDate=datetime();
	if (tokenIgnore=0 & (missing(prevTokenDate) | prevTokenDate<&conceptTokenLatest)) then do;
		conceptTokenHash=conceptTokenIter.first();
		do while (conceptTokenHash=0);
			if (missing(prevTokenDate) | prevTokenDate<conceptTokenDate) then do;
				if (prxmatch(conceptTokenRegex,strip(token))) then do;
					output _tokenMatches;
				end;
			end;
			conceptTokenHash=conceptTokenIter.next();
		end;
	end;
	output &dataTokenized;
	logPctDone=(_N_/logCount)*100;
	if (logPctDone>=(logLastPct+1) & logPctDone<100) then do;
		logLastPct=round((_N_/logCount)*100,1.0);
		logTimePass=datetime()-logStart;
		log_n=put(_N_,8.0);
		if (logTimePass<=60) then log_text=strip(put(_N_,COMMA10.0)) || " of " || strip(put(logCount,COMMA10.0)) || " (" || strip(put(logLastPct,8.0)) || "%)" || " in " || strip(put(logTimePass,4.1)) || " seconds.";
		else log_text=strip(put(_N_,COMMA10.0)) || " of " || strip(put(logCount,COMMA8.0)) || " (" || strip(put(logLastPct,8.0)) || "%)" || " in " || strip(put(logTimePass/60,5.1)) || " minutes.";
		put "NOTE: Finished matching tokens for observation #" log_text;
	end;
run;
data &dataTokenMatches(compress=yes drop=tokenIgnore);
	set %if (%sysfunc(exist(&dataTokenMatches))>0) %then &dataTokenMatches; _tokenMatches;;
run;
proc sql;
	drop table conceptTokenRegex, _tokenMatches;
quit;
%mend matchTokens;
