/***************************************************************************/
/*Name:          SCENT: Dictionary                                         */
/*Location:      Undefined                                                 */
/*Author:        Justin A. Strauss                                         */
/*Organization:  Kaiser Permanente Southern California                     */
/*Purpose:       Create concept dictionary                                 */
/*Modified:                                                                */
/***************************************************************************/

%include "&scentDict\Source\createDictionary_Compile.sas";

/*
%createDictionary(
	concepts="F:\Macros\PathAlgorithm\createDictionary_Concepts.xls"
	,cleaning="F:\Macros\PathAlgorithm\createDictionary_Patterns_Clean.xls"
	,enhancements="F:\Macros\PathAlgorithm\createDictionary_Patterns_Enhance.xls"
	,out="work.createDictionary"
);
*/

%macro createDictionary(concepts=,cleaning=,enhancements=,out="work.createDictionary",debug="NO");

/*remove quotation marks surrounding supplied parameters*/
%if (%length(&out)>0) %then %let out=%sysfunc(dequote(&out));
%if (%length(&concepts)>0) %then %let concepts=%sysfunc(dequote(&concepts));
%if (%length(&cleaning)>0) %then %let cleaning=%sysfunc(dequote(&cleaning));
%if (%length(&enhancements)>0) %then %let enhancements=%sysfunc(dequote(&enhancements));
%if (%length(&debug)>0) %then %let debug=%sysfunc(dequote(&debug));

	%macro GetSysOpts;
	%local ls notes source mprint mlogic implmac;
	%let ls=%sysfunc(getoption(ls,keyword));
	%let notes=%sysfunc(getoption(notes,keyword));
	%let source=%sysfunc(getoption(source,keyword));
	%let mprint=%sysfunc(getoption(mprint,keyword));
	%let mlogic=%sysfunc(getoption(mlogic,keyword));
	%let implmac=%sysfunc(getoption(implmac,keyword));
	&ls &notes &source &mprint &mlogic
	%mend GetSysOpts;

/*store and set global options*/
%local opts;
%let opts=%GetSysOpts;
%if (%upcase(&debug)~=NO) %then %do;
	options mprint mlogic source;
%end; %else %do;
	options nomprint nomlogic nosource;
%end;

	%macro CleanDescription(data=,out=,workbook=,text=,cleaned=);
	%if (%length(&data)>0) %then %let data=%sysfunc(dequote(&data));
	%if (%length(&out)>0) %then %let out=%sysfunc(dequote(&out));
	%else %let out=&data;
	%if (%length(&workbook)>0) %then %let workbook=%sysfunc(dequote(&workbook));
	%if (%length(&text)>0) %then %let text=%sysfunc(dequote(&text));
	%if (%length(&cleaned)>0) %then %let cleaned=%sysfunc(dequote(&cleaned));
	%else %let cleaned=&text;

	/*read in the regular expression patterns used in preprocessing*/
	libname xls "&workbook";
	    data patterns;
        	set xls."SAS$"n(rename=(pattern=patternText enabled=patternEnabled));
			length pattern 8;
			call missing(pattern);
    	run;
	libname xls clear;

	data &out(drop=_:);
		set &data;
		if (_N_=1) then do;
			/*read regular expression patterns into hash table*/
			length patternID pattern patternEnabled 8 patternText $500;
			drop patternID pattern patternEnabled patternText;
			declare hash patterns(dataset:"patterns", ordered:"a");
			declare hiter hlookup('patterns');
			call patterns.defineKey('patternID');
			call patterns.defineData('patternID');
			call patterns.defineData('patternEnabled');
			call patterns.defineData('patternText');
			call patterns.defineData('pattern');
			call patterns.defineDone();
			call missing(patternID, pattern, patternEnabled, patternText);
			_rc=hlookup.first();
			do while (_rc=0 | _remove);
				if (_remove) then do;
					call patterns.remove(key: _removeID);
					if (_rc~=0) then leave;
				end;
				_remove=ifn(patternEnabled,0,1,0);
				_removeID=ifn(patternEnabled,.,patternID,.);
				if (patternEnabled) then call patterns.replace(key: patternID, data: patternID, data: patternEnabled, data: patternText, data: prxparse("S" || strip(patternText)));
				_rc=hlookup.next();
			end;
		end;
		%if (&cleaned~=&text) %then %do;
			&cleaned=&text;
		%end;

		_rc=hlookup.first();
		do while (_rc=0);
			&cleaned=prxchange(pattern, -1, &cleaned);
			_rc=hlookup.next();
		end;
	run;
	proc sql;
		drop table patterns;
	quit;
	%mend CleanDescription;

	%macro EnhanceTokens(data=,out=,workbook=,tokens=);
	%if (%length(&data)>0) %then %let data=%sysfunc(dequote(&data));
	%if (%length(&out)>0) %then %let out=%sysfunc(dequote(&out));
	%else %let out=&data;
	%if (%length(&workbook)>0) %then %let workbook=%sysfunc(dequote(&workbook));
	%if (%length(&tokens)>0) %then %let tokens=%sysfunc(dequote(&tokens));

	/*read in the regular expression patterns used in preprocessing*/
	libname xls "&workbook";
	    data patterns;
        	set xls."SAS$"n(rename=(pattern=patternText enabled=patternEnabled));
			length pattern 8;
			call missing(pattern);
    	run;
	libname xls clear;

	data &out(drop=_tokenCount);
		set &data;
		if (_N_=1) then do;
			/*read regular expression patterns into hash table*/
			length patternID pattern patternEnabled 8 patternText $500;
			drop patternID pattern patternEnabled patternText;
			declare hash patterns(dataset:"patterns", ordered:"a");
			declare hiter hlookup('patterns');
			call patterns.defineKey('patternID');
			call patterns.defineData('patternID');
			call patterns.defineData('patternEnabled');
			call patterns.defineData('patternText');
			call patterns.defineData('pattern');
			call patterns.defineDone();
			call missing(patternID, pattern, patternEnabled, patternText);
			_rc=hlookup.first();
			do while (_rc=0 | _remove);
				if (_remove) then do;
					call patterns.remove(key: _removeID);
					if (_rc~=0) then leave;
				end;
				_remove=ifn(patternEnabled,0,1,0);
				_removeID=ifn(patternEnabled,.,patternID,.);
				if (patternEnabled) then call patterns.replace(key: patternID, data: patternID, data: patternEnabled, data: patternText, data: prxparse("S" || strip(patternText)));
				_rc=hlookup.next();
			end;
		end;
		array &tokens{50} $200;
		do _tokenCount=1 to dim(&tokens);
			if (lengthn(&tokens{_tokenCount})=0) then continue;
			_rc=hlookup.first();
			do while (_rc=0);
				&tokens{_tokenCount}=prxchange(pattern, -1, &tokens{_tokenCount});
				_rc=hlookup.next();
			end;

		end;
	run;
	proc sql;
		drop table patterns;
	quit;
	%mend EnhanceTokens;

	%macro Tokenize(data=,out=,text=,tokens="conceptToken");
	%if (%length(&data)>0) %then %let data=%sysfunc(dequote(&data));
	%if (%length(&out)>0) %then %let out=%sysfunc(dequote(&out));
	%else %let out=&data;
	%if (%length(&text)>0) %then %let text=%sysfunc(dequote(&text));
	%if (%length(&tokens)>0) %then %let tokens=%sysfunc(dequote(&tokens));
	data &out(drop=_:);
		set &data;
			array &tokens{50} $200;
			if (_N_=1) then do;
				_tokenRegex=prxparse("/(?<=\b)([0-9A-Z\-\_]+)(?=\b)/i");
				retain _tokenRegex;
			end;
			_tokenCount=0; _tokenStart=1; _tokenStop=lengthn(&text);
			if (_tokenStop=0) then return;
			call prxnext(_tokenRegex, _tokenStart, _tokenStop, &text, _tokenPos, _tokenLen);
			do while (_tokenPos>0 & _tokenLen>0);
				if (lengthn(substr(&text, _tokenPos, _tokenLen))) then do;
					_tokenCount=_tokenCount+1;
					&tokens{_tokenCount}=upcase(substr(&text, _tokenPos, _tokenLen));
				end;
				call prxnext(_tokenRegex, _tokenStart, _tokenStop, &text, _tokenPos, _tokenLen);
			end;
	run;
	%mend Tokenize;

	%macro MergeSiblings(data=,out=,tokens=,code="code",newcode="newcode");
	%if (%length(&data)>0) %then %let data=%sysfunc(dequote(&data));
	%if (%length(&out)>0) %then %let out=%sysfunc(dequote(&out));
	%else %let out=&data;
	%if (%length(&tokens)>0) %then %let tokens=%sysfunc(dequote(&tokens));
	%if (%length(&code)>0) %then %let code=%sysfunc(dequote(&code));
	%if (%length(&newcode)>0) %then %let newcode=%sysfunc(dequote(&newcode));
	data &out(drop= _tokenCount);
		set &data;
		array &tokens{50} $200;
		do _tokenCount=1 to dim(&tokens);
			if (lengthn(&tokens{_tokenCount})=0) then continue;
			if (substr(&code,lengthn(&code),1)="2" & prxmatch("/(?i)(?<=\b)(SITU|NON.?INVASIVE|NON.?INFILTRATING|INTRAEPITHELIAL|DCIS)(?=\b)/",&tokens{_tokenCount})) then do;
				&tokens{_tokenCount}=prxchange("s/(?i)(?<=\b)(SITU|NON.?INVASIVE|NON.?INFILTRATING|INTRAEPITHELIAL|DCIS)(?=\b)//",-1,&tokens{_tokenCount});
				&newcode=substr(&code,1,lengthn(&code)-1) || "X";
			end; else if (substr(&code,lengthn(&code),1)="6" & prxmatch("/(?i)(?<=\b)METASTA(SES|SIS|TIC|SISING)(?=\b)/",&tokens{_tokenCount})) then do;
				&tokens{_tokenCount}=prxchange("s/(?i)(?<=\b)METASTA(SES|SIS|TIC|SISING)(?=\b)//",-1,&tokens{_tokenCount});
				&newcode=substr(&code,1,lengthn(&code)-1) || "X";
			end; else if (substr(&code,lengthn(&code),1)="1" & prxmatch("/(?i)(?<=\b)BORDERLINE( MALIGNANCY)?(?=\b)/",&tokens{_tokenCount})) then do;
				&tokens{_tokenCount}=prxchange("s/(?i)(?<=\b)BORDERLINE( MALIGNANCY)?(?=\b)//",-1,&tokens{_tokenCount});
				&newcode=substr(&code,1,lengthn(&code)-1) || "X";
			end; else if (substr(&code,lengthn(&code),1)="3" & prxmatch("/(?i)(?<=\b)INVASIVE(?=\b)/",&tokens{_tokenCount})) then do;
				&tokens{_tokenCount}=prxchange("s/(?i)(?<=\b)INVASIVE(?=\b)//",-1,&tokens{_tokenCount});
				&newcode=substr(&code,1,lengthn(&code)-1) || "X";				
			end;
		end;
	run;
	%mend MergeSiblings;

	%macro Pluralize(data=,out=,tokens=);
	%if (%length(&data)>0) %then %let data=%sysfunc(dequote(&data));
	%if (%length(&out)>0) %then %let out=%sysfunc(dequote(&out));
	%else %let out=&data;
	%if (%length(&tokens)>0) %then %let tokens=%sysfunc(dequote(&tokens));
	data &out(drop=_tokenCount _suffixCount);
		set &data;
		array &tokens{50} $200;
		if (_N_=1) then do;
			array _suffix {11,2} $50 _temporary_ (
				"SIS"	"S(?:IS|ES|IC)"
				"MAN"	"M(?:AN|EN)"
				"SH"	"SH(?:ES)?"
				"CH"	"CH(?:ES)?"
				"FF"	"(?:FF|VES)"
				"Y"		"(?:Y|IES)"
				"Z"		"Z(?:ES)?"
				"X"		"X(?:ES)?"
				"S"		"S(?:ES)?"
				"A"		"A[LS]?"
				""		"S?"
			);
			retain _suffix:;
		end;
		do _tokenCount=1 to dim(&tokens);
			if (lengthn(&tokens{_tokenCount})=0) then continue;
			if (lengthn(&tokens{_tokenCount})>=4 & prxmatch("/(?i)[A-Z](?=\z)/",strip(&tokens{_tokenCount}))) then do;
				do _suffixCount=1 to dim(_suffix,1);
					_stem=substr(&tokens{_tokenCount},1,lengthn(&tokens{_tokenCount})-lengthn(_suffix{_suffixCount,1}));
					if (lengthn(_suffix{_suffixCount,1})) then do;
						_suffixCurrent=substr(&tokens{_tokenCount},lengthn(&tokens{_tokenCount})-lengthn(_suffix{_suffixCount,1})+1,lengthn(_suffix{_suffixCount,1}));
						if (_suffixCurrent=upcase(_suffix{_suffixCount,1})) then do;
							&tokens{_tokenCount}=strip(_stem) || upcase(strip(_suffix{_suffixCount,2}));
							leave;
						end;
					end; else do;
						&tokens{_tokenCount}=strip(_stem) || upcase(strip(_suffix{_suffixCount,2}));
						leave;
					end;
				end;
			end;
		end;
	run;
	%mend Pluralize;

	%macro CountTokens(data=,out=,tokens=, table="tokencounts", count=);
	%if (%length(&data)>0) %then %let data=%sysfunc(dequote(&data));
	%if (%length(&out)>0) %then %let out=%sysfunc(dequote(&out));
	%else %let out=&data;
	%if (%length(&tokens)>0) %then %let tokens=%sysfunc(dequote(&tokens));
	%if (%length(&table)>0) %then %let table=%sysfunc(dequote(&table));
	%if (%length(&count)>0) %then %let count=%sysfunc(dequote(&count));
	data &out(drop=_:);
		set &data nobs=lastobs;
			array &tokens{50} $200;
			if (_N_=1) then do;
				length _tokenText $100 _tokenTextCount 8;
				%if (%length(&count)>0) %then %do;
					length &count 8;
				%end;
				declare hash hwords();
				call hwords.defineKey('_tokenText');
				call hwords.defineData('_tokenText');
				call hwords.defineData('_tokenTextCount');
				call hwords.defineDone();
				call missing(_tokenText, _tokenTextCount);
			end;
			%if (%length(&count)>0) %then %do;
				&count=0;
			%end;
			do _tokenCount=1 to dim(&tokens);
				if (lengthn(&tokens{_tokenCount})) then do;
					%if (%length(&count)>0) %then %do;
						&count=&count+1;
					%end;
					if (hwords.find(key: &tokens{_tokenCount})=0) then do;
						call hwords.replace(key: &tokens{_tokenCount}, data: &tokens{_tokenCount}, data: _tokenTextCount+1);
					end; else do;
						call hwords.add(key: &tokens{_tokenCount}, data: &tokens{_tokenCount}, data: 1);
					end;
				end;	
			end;
			if (_N_=lastobs) then do;
				call hwords.output(dataset: "work.&table");
			end;
	run;
	%mend CountTokens;

	%macro SortTokens(data=,out=,tokens=);
	%if (%length(&data)>0) %then %let data=%sysfunc(dequote(&data));
	%if (%length(&out)>0) %then %let out=%sysfunc(dequote(&out));
	%else %let out=&data;
	%if (%length(&tokens)>0) %then %let tokens=%sysfunc(dequote(&tokens));
	data &out(drop=_:);
		set &data;
		array &tokens{50} $200;
		array _tokenCounts{25} 8;
		if (_N_=1) then do;
			length _tokenText $100 _tokenTextCount 8;
			declare hash hwords(dataset: 'tokencounts');
			call hwords.defineKey('_tokenText');
			call hwords.defineData('_tokenText');
			call hwords.defineData('_tokenTextCount');
			call hwords.defineDone();
			call missing(_tokenText, _tokenTextCount);
		end;
		do _tokenCount=1 to dim(&tokens);
			do _tokenCountIinner=1 to dim(&tokens);
				if (lengthn(&tokens{_tokenCountIinner})=0) then continue;
				else if (_tokenCount~=_tokenCountIinner & &tokens{_tokenCountIinner}=&tokens{_tokenCount}) then &tokens{_tokenCountIinner}='';
			end;
		end;
		do _tokenCount=1 to dim(&tokens);
			if (lengthn(&tokens{_tokenCount})=0) then continue;
			if (hwords.find(key: &tokens{_tokenCount})=0) then do;
				_tokenCounts{_tokenCount}=_tokenTextCount;
			end;
		end;
		do until (_sorted);
			_sorted=1;
			do _tokenCount = dim(_tokenCounts) to 2 by -1;
				if ((lengthn(&tokens{_tokenCount})>0 & lengthn(&tokens{_tokenCount-1})=0) | (_tokenCounts{_tokenCount}>_tokenCounts{_tokenCount-1})) then do;
					_temp=&tokens{_tokenCount-1};
					_tempCount=_tokenCounts{_tokenCount-1};
					&tokens{_tokenCount-1}=&tokens{_tokenCount};
					_tokenCounts{_tokenCount-1}=_tokenCounts{_tokenCount};
					&tokens{_tokenCount}=_temp;
					_tokenCounts{_tokenCount}=_tempCount;
					_sorted=0;
				end;
			end;
		end;
	run;
	%mend SortTokens;

	%macro CleanDictionary(data=,out=,tokens=);
	%if (%length(&data)>0) %then %let data=%sysfunc(dequote(&data));
	%if (%length(&out)>0) %then %let out=%sysfunc(dequote(&out));
	%else %let out=&data;
	%if (%length(&tokens)>0) %then %let tokens=%sysfunc(dequote(&tokens));
	data &out(drop=_:);
		set &data nobs=lastobs;
			array &tokens{50} $200;
			retain _tokenMax;
			do _tokenCount=1 to dim(&tokens);
				if (lengthn(&tokens{_tokenCount})) then do;
					if (missing(_tokenMax) | _tokenCount>_tokenMax) then _tokenMax=_tokenCount;
				end;
			end;
			if (_N_=lastobs) then do;
				call symputx("droptokens","&tokens" || strip(put(_tokenMax+1,8.0)) || "-&tokens" || strip(put(dim(&tokens),8.0)),"G");
			end;
	run;
	proc sort data=&out out=&out(drop=&droptokens); by codeClass &tokens.1 descending tokenCount &tokens:; run;
	%mend CleanDictionary;

	%macro MergeSiblingCodes(data=,out=,tokens=,code="code",newcode="newcode");
	%if (%length(&data)>0) %then %let data=%sysfunc(dequote(&data));
	%if (%length(&out)>0) %then %let out=%sysfunc(dequote(&out));
	%else %let out=&data;
	%if (%length(&tokens)>0) %then %let tokens=%sysfunc(dequote(&tokens));
	%if (%length(&code)>0) %then %let code=%sysfunc(dequote(&code));
	%if (%length(&newcode)>0) %then %let newcode=%sysfunc(dequote(&newcode));
	data _null_;
		set &data(obs=1);
		array &tokens{*} &tokens:;
		call symputx("tokenmax","&tokens" || strip(put(dim(&tokens),8.0)),"G");
	run;
	%if (%length(&data)>0) %then %let data=%sysfunc(dequote(&data));
	%if (%length(&out)>0) %then %let out=%sysfunc(dequote(&out));
	%else %let out=&data;
	%if (%length(&tokens)>0) %then %let tokens=%sysfunc(dequote(&tokens));
	%if (%length(&code)>0) %then %let code=%sysfunc(dequote(&code));
	%if (%length(&newcode)>0) %then %let newcode=%sysfunc(dequote(&newcode));
	proc sort data=&data out=&out; by codeClass &tokens.1 descending tokenCount &tokens: descending &newcode; run;
	data &out(drop=_: codeDesc codeDescExclude &newcode);
		length codeSystem codeClass $25 Code $10 codeMorph $20;
		format codeSystem codeClass $25. Code $10. codeMorph $20.;
		informat codeSystem codeClass $25. Code $10. codeMorph $20.;
		set &out(rename=(&code=_code codeMorph=_codeMorph));
		by codeClass &tokens.1 descending tokenCount &tokens: descending &newcode;
		retain code codeMorph;
		if (first.&tokenmax) then do;
			if (lengthn(&newcode) & substr(_code,1,lengthn(_code)-1)=substr(&newcode,1,lengthn(&newcode)-1)) then do;
				code=&newcode;
				codeMorph="Variable";
			end; else do;
				code=_code;
				codeMorph=_codeMorph;
			end;
		end;
	run;
	proc sort data=&out out=&out(drop=tokenCount) nodupkey; by codeClass &tokens.1 descending tokenCount &tokens: code; run;
	%mend MergeSiblingCodes;

/*import createDictionary spreadsheet*/
libname xls "&concepts";
data _tempdict(drop=codeMorph_:);
	length codeSystem codeClass $25 Code $10 codeDesc codeDescExclude $500 codeMorph codeSite $20;
	format codeSystem codeClass $25. Code $10. codeDesc codeDescExclude $500. codeMorph codeSite $20.;
	informat codeSystem codeClass $25. Code $10. codeDesc codeDescExclude $500. codeMorph codeSite $20.;
	set xls."SAS$"n;
run;
libname xls clear;

/*clean descriptions*/
%CleanDescription(data="_tempdict", workbook="&cleaning", text="codeDesc");
%CleanDescription(data="_tempdict", workbook="&cleaning", text="codeDescExclude");

/*tokenize descriptions into words*/
%Tokenize(data="_tempdict", text="codeDesc", tokens="conceptToken");
%Tokenize(data="_tempdict", text="codeDescExclude", tokens="excludeToken");

/*merge sibling concepts differing only by tracked disease extents*/
%MergeSiblings(data="_tempdict", tokens="conceptToken", code="code", newcode="newcode");

/*enhance tokens with regular expressions*/
%EnhanceTokens(data="_tempdict", workbook="&enhancements", tokens="conceptToken");
%EnhanceTokens(data="_tempdict", workbook="&enhancements", tokens="excludeToken");

/*add optional plural suffixes to tokens*/
%Pluralize(data="_tempdict", tokens="conceptToken");
%Pluralize(data="_tempdict", tokens="excludeToken");

/*count number of times each word is used, sort by token count, and clean unused token slots*/
%CountTokens(data="_tempdict", tokens="conceptToken", table="tokencounts", count="tokenCount");
%SortTokens(data="_tempdict", tokens="conceptToken");
%CleanDictionary(data="_tempdict", tokens="conceptToken");

/*count number of times each word is used, sort by token count, and clean unused token slots*/
%CountTokens(data="_tempdict", tokens="excludeToken", table="excludecounts");
%SortTokens(data="_tempdict", tokens="excludeToken");
%CleanDictionary(data="_tempdict", tokens="excludeToken");

/*finalize sibling concept merge and deduplicate &out*/
%MergeSiblingCodes(data="_tempdict",tokens="conceptToken",code="code",newcode="newcode");

data _tempdict;
	length conceptID conceptDate 8;
	format conceptDate datetime16.0;
	set _tempdict;
	conceptID=_N_;
	conceptDate=datetime();
run;

proc sql noprint;
	create table &out as select * from _tempdict;
	drop table _tempdict, tokencounts, excludecounts;
quit;

/*restore original global options*/
options &opts;

%mend createDictionary;
