/***************************************************************************/
/*Name:          SCENT: Dictionary                                         */
/*Location:      Undefined                                                 */
/*Author:        Justin A. Strauss                                         */
/*Organization:  Kaiser Permanente Southern California                     */
/*Purpose:       Tokenize concept dictionary                               */
/*Modified:                                                                */
/***************************************************************************/

%macro createDictionary_Compile;
data conceptTokenRegex(keep=_conceptToken conceptDate rename=(_conceptToken=conceptToken));
	length _conceptToken $200;
	set &dataDict;
	array conceptToken {*} conceptToken:;
	array excludeToken {*} excludeToken:;
	do _i=1 to dim(conceptToken);
		if (lengthn(conceptToken{_i})) then do;
			_conceptToken=conceptToken{_i};
			output;
		end;
	end;
	do _i=1 to dim(excludeToken);
		if (lengthn(excludeToken{_i})) then do;
			_conceptToken=excludeToken{_i};
			output;
		end;
	end;
run;
proc sort data=conceptTokenRegex nodupkey; by conceptToken; run;

%if (%sysfunc(exist(_dict.conceptTokenRegex))>0) %then %do;
	proc sql noprint;
		select max(conceptTokenID) into :conceptTokenMax from _dict.conceptTokenRegex;
	quit;
	proc sort data=_dict.conceptTokenRegex; by conceptToken; run;
	data _dict.conceptTokenRegex(keep=conceptTokenID conceptTokenDate conceptTokenActive conceptToken);
		length conceptTokenID newTokens conceptTokenDate conceptTokenActive 8 conceptToken $200;
		format conceptTokenDate datetime16.0;
		merge _dict.conceptTokenRegex(in=a rename=(conceptTokenDate=oldDate)) conceptTokenRegex(in=b);
		by conceptToken;
		if (a=1 & b=1) then do;
			conceptTokenDate=oldDate;
			conceptTokenActive=1;
		end; else if (b=1) then do;
			newTokens+1;
			conceptTokenID=&conceptTokenMax+newTokens;
			conceptTokenDate=conceptDate;
			conceptTokenActive=1;
		end; else if (a=1) then do;
			conceptTokenDate=oldDate;
			conceptTokenActive=0;
		end;
		output;
	run;
%end; %else %do;
	data _dict.conceptTokenRegex(keep=conceptTokenID conceptTokenDate conceptTokenActive conceptToken);
		length conceptTokenID conceptTokenDate conceptTokenActive 8 conceptToken $200;
		format conceptTokenDate datetime16.0;
		set conceptTokenRegex;
		conceptTokenID=_N_;
		conceptTokenDate=conceptDate;
		conceptTokenActive=1;
	run;
%end;

data temp(keep=conceptCode conceptDate _conceptTokenSign _conceptToken conceptSearchDate rename=(_conceptTokenNum=conceptTokenNum _conceptTokenSign=conceptTokenSign _conceptToken=conceptToken));
	length _conceptCode $10 _conceptTokenNum _conceptTokenSign 8 _conceptToken $200 conceptSearchDate 8;
	format conceptSearchDate datetime16.0;
	call missing(conceptSearchDate);
	set &dataDict;
	array conceptToken {*} conceptToken:;
	array excludeToken {*} excludeToken:;
	do _i=1 to dim(conceptToken);
		if (lengthn(conceptToken{_i})) then do;
			conceptCode=strip(code);
			_conceptTokenNum=_i;
			_conceptTokenSign=1;
			_conceptToken=conceptToken{_i};
			output;
		end;
	end;
	do _i=1 to dim(excludeToken);
		if (lengthn(excludeToken{_i})) then do;
			conceptCode=strip(code);
			_conceptTokenNum=_i;
			_conceptTokenSign=-1;
			_conceptToken=excludeToken{_i};
			output;
		end;
	end;
run;
proc sql;
	create table _dict.conceptTokens(compress=yes) as
		select a.*, b.conceptTokenID
		from temp as a left join _dict.conceptTokenRegex as b
		on (a.conceptToken=b.conceptToken)
		order by a.conceptCode, a.conceptTokenSign desc, a.conceptTokenNum;
	create table check as select * from temp;
	drop table temp, conceptTokenRegex;
quit;

/*remove matches to concept tokens that no longer exist*/
proc sql;
	%if (%sysfunc(exist(&dataMatched))>0) %then %do;
		create table temp(compress=yes) as
			select a.*
			from &dataMatched as a left join _dict.conceptTokenRegex as b
			on (a.conceptTokenID=b.conceptTokenID)
			where (b.conceptTokenActive=1);
		create table &dataMatched(compress=yes) as select * from temp;
		drop table temp;
	%end;
	delete from _dict.conceptTokenRegex where (conceptTokenActive~=1);
	alter table _dict.conceptTokenRegex drop column conceptTokenActive;
quit;

%mend createDictionary_Compile;