/***************************************************************************/
/*Name:          SCENT: Dictionary                                         */
/*Location:      Undefined                                                 */
/*Author:        Justin A. Strauss                                         */
/*Organization:  Kaiser Permanente Southern California                     */
/*Purpose:       Tokenize text                                             */
/*Modified:                                                                */
/***************************************************************************/

%macro preprocessText_Tokenize;
data &dataTokenized(compress=yes keep=reportKey lineNum linePos tokenDate tokenIgnore tokenNum tokenPos tokenLen token);
	length tokenKey $50;
	length reportKey $50 lineNum linePos tokenDate tokenNum tokenIgnore 8 tokenPos tokenLen 8 token $200;
	format tokenDate datetime16.0;
	call missing(tokenDate);
	set &dataPreproc nobs=logCount;

	retain logLastPct logStart;
	if (_N_=1) then do;
		logStart=datetime();
		logLastPct=0;
		/***CREATE PATTERNS TO PARSE REPORT LINES AND LINE WORDS***/
		lineRegex=prxparse('/(?i)(?:(?<=\A)|(?<=<BR>))(.*?)(?:(?=\z)|(?=<BR>))/');
		wordRegex=prxparse('/(?i)(?<=\b)((?:[0-9]+[\.\/\-][0-9]+)|(?:[0-9A-Z\-\_]+))(?=\b)/');
		retain lineRegex wordRegex;
	end;

	reportKey=link_value;

	lineNum=0;
	lineStart=1;
	lineStop=lengthn(final_preproc);
	if (lineStop=0) then return;
	call prxnext(lineRegex, lineStart, lineStop, final_preproc, linePos, lineLen);
	do while (linePos>0 & lineLen>0);
		lineNum=lineNum+1;
		line=substr(final_preproc, linePos, lineLen);
		/*match tokens and store in separate hash*/
		tokenNum=0;
		tokenStart=1;
		tokenStop=lengthn(line);
		if (tokenStop=0) then return;
		call prxnext(wordRegex, tokenStart, tokenStop, line, tokenPos, tokenLen);
		do while (tokenPos>0 & tokenLen>0);
			tokenNum=tokenNum+1;
			tokenKey=strip(reportKey) || "-" || put(lineNum,z4.) || "-" || put(tokenNum,z4.);
			token=substr(line, tokenPos, tokenLen);
			tokenIgnore=0;
			if (prxmatch("/(?i)\A(?:A[STN]?|ALSO|AND|BY|FOR|FROM|HAS|I[NST]|LIKE|O[FR]|OUT|THAT|THE|THIS|TO|WITH|ARE|ON|WAS|BE)\z/",strip(token))) then tokenIgnore=1;
			else if (prxmatch("/(?i)\A(?:WERE|THERE|THEIR)\z/",strip(token))) then tokenIgnore=1;
			else if (prxmatch("/(?i)\A(?:AM|PM|CM|MM|DATE|TIME|MD|GT|LT|X)\z/",strip(token))) then tokenIgnore=1;
			else if (prxmatch("/(?i)\A(?:[0-9]+[\.\/\-][0-9]+)\z/",strip(token))) then tokenIgnore=1;
			else if (prxmatch("/(?i)\A(?:[0-9]+)\z/",strip(token))) then tokenIgnore=1;
			else if (prxmatch("/(?i)\A(?:[[:punct:]]+)\z/",strip(token))) then tokenIgnore=1;
			output &dataTokenized;
			call prxnext(wordRegex, tokenStart, tokenStop, line, tokenPos, tokenLen);
		end;
		call prxnext(lineRegex, lineStart, lineStop, final_preproc, linePos, lineLen);
	end;
	logPctDone=(_N_/logCount)*100;
	if (logPctDone>=(logLastPct+1) & logPctDone<100) then do;
		logLastPct=round((_N_/logCount)*100,1.0);
		logTimePass=datetime()-logStart;
		log_n=put(_N_,8.0);
		if (logTimePass<=60) then log_text=strip(put(_N_,COMMA10.0)) || " of " || strip(put(logCount,COMMA10.0)) || " (" || strip(put(logLastPct,8.0)) || "%)" || " in " || strip(put(logTimePass,4.1)) || " seconds.";
		else log_text=strip(put(_N_,COMMA8.0)) || " of " || strip(put(logCount,COMMA10.0)) || " (" || strip(put(logLastPct,8.0)) || "%)" || " in " || strip(put(logTimePass/60,5.1)) || " minutes.";
		put "NOTE: Finished tokenizing observation #" log_text;
	end;
run;
%mend preprocessText_Tokenize;
