/***************************************************************************/
/*Name:          SCENT, Matching, Coding, & Report Classification          */
/*Location:      Undefined                                                 */
/*Author:        Justin A. Strauss                                         */
/*Purpose:                                                                 */
/*Date:                                                                    */
/*Modified:                                                                */
/*License: Licensed under the Apache License, Version 2.0 (the License).  You may not use this file except in compliance with the License.  You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0 
  
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an AS IS BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the specific language governing permissions and limitations under the License. 
*/
/***************************************************************************/

%macro scent_coder(
	data=
	,var=preprocessed
	,dict=core.scent_lookup
	,defs="path\scent_dictionaryconcepts.xls"
	,sites="path\scent_topology.xls"
	,obs=
	,out=
	,cancer=BCa
	);

	%macro GetSysOpts;
	%local ls notes source mprint mlogic implmac;
	%let ls=%sysfunc(getoption(ls,keyword));
	%let notes=%sysfunc(getoption(notes,keyword));
	%let source=%sysfunc(getoption(source,keyword));
	%let mprint=%sysfunc(getoption(mprint,keyword));
	%let mlogic=%sysfunc(getoption(mlogic,keyword));
	%let implmac=%sysfunc(getoption(implmac,keyword));
	&ls &notes &source &mprint &mlogic
	%mend GetSysOpts;

%local opts;
%let opts=%GetSysOpts;
options nomprint nomlogic notes nosource;

%if (%length(&data)>0) %then %let data=%sysfunc(dequote(&data));
%else %return;
%if (%length(&var)>0) %then %let var=%sysfunc(dequote(&var));
%if (%length(&obs)>0) %then %let obs=%sysfunc(dequote(&obs));
%if (%length(&out)>0) %then %let out=%sysfunc(dequote(&out));
%else %let out=&data;
%if (%length(&dict)>0) %then %let dict=%sysfunc(dequote(&dict));
%if (%length(&defs)>0) %then %let defs=%sysfunc(dequote(&defs));
%if (%length(&sites)>0) %then %let sites=%sysfunc(dequote(&sites));
%if (%length(&cancer)>0) %then %let cancer=%sysfunc(dequote(&cancer));
data lookup_scent(compress=yes);
	set &dict;
run;
proc sort data=lookup_scent; by snm_type descending txt_pos4 descending txt_pos3 descending txt_pos2 descending txt_pos1; run;

data inputset(compress=yes);
	set &data;
	%if (%length(&obs)>0) %then %do;
		if (_N_<=&obs);
	%end;
run;

data temp(compress=yes drop=&var regex_: log_: line _: txt_pos: txt_neg: snm_code snm_type snm_class);
	length line _lineforarray $1000 _newline $2000;
	length regex_start regex_stop regex_position regex_length 8;
	array _line_array{200} $100;
	array _skip_array{200} 8;
	array _word_pos{10,10} 8;
	set inputset nobs=log_count;
	length coded $10000 nlp_codes_all nlp_codes $2000 nlp_subtypes nlp_subclasses nlp_pvals $1000;
	length nlp_match_ajcc 8 nlp_ajcc $25 nlp_ajcc_t nlp_ajcc_n nlp_ajcc_m $5;
%if (%upcase(&cancer)=PCA) %then %do;
	length nlp_match_gleason nlp_gleason_one nlp_gleason_two nlp_gleason 8;
%end;
	regex_line=prxparse('/(?i)(?:\A|<BR>)(.*?)(?=<BR>|\z)/');
	retain log_lastpct log_start;
	if (_N_=1) then do;
		log_start=datetime();
		log_lastpct=0;

		length _hkey $100 _hpos 8;
		declare hash h(multidata: 'yes');
		call h.defineKey('_hkey');
		call h.defineData('_hpos');
		call h.defineDone();
		call missing(_hkey, _hpos);

		length _mkey _mval $100 _msub _mclass $1 _mpval $5 _mpos 8;
		declare hash hmatch(multidata: 'yes', ordered: 'd');
		call hmatch.defineKey('_mkey');
		call hmatch.defineData('_mpos');
		call hmatch.defineData('_mval');
		call hmatch.defineData('_msub');
		call hmatch.defineData('_mclass');
		call hmatch.defineData('_mpval');
		call hmatch.defineDone();
		call missing(_mkey, _mpos, _mval, _msub, _mclass, _mpval);
		declare hiter miter('hmatch');

		length _tkey $3 _taken 8;
		declare hash htaken();
		call htaken.defineKey('_tkey');
		call htaken.defineData('_taken');
		call htaken.defineDone();
		call missing(_tkey, _taken);

		length snm_code $7 snm_type snm_class $10 txt_pos1 txt_pos2 txt_pos3 txt_pos4 txt_pos5 txt_pos6 txt_neg1 $100;
		declare hash hlookup(dataset:"lookup_scent");
		call hlookup.defineKey('snm_code');
		call hlookup.defineData('snm_code');
		call hlookup.defineData('snm_type');
		call hlookup.defineData('snm_class');
		call hlookup.defineData('txt_pos1');
		call hlookup.defineData('txt_pos2');
		call hlookup.defineData('txt_pos3');
		call hlookup.defineData('txt_pos4');
		call hlookup.defineData('txt_pos5');
		call hlookup.defineData('txt_pos6');
		call hlookup.defineData('txt_neg1');
		call hlookup.defineDone();
		call missing(snm_code, snm_type, snm_class, txt_pos1, txt_pos2, txt_pos3, txt_pos4, txt_pos5, txt_pos6, txt_neg1);
		declare hiter iter('hlookup');

		length _neg_key $2 _negate_pattern $200 _negate_maxbefore _negate_maxafter 8;
		declare hash hnegate(multidata: 'yes');
		call hnegate.defineKey('_neg_key');
		call hnegate.defineData('_negate_pattern');
		call hnegate.defineData('_negate_maxbefore');
		call hnegate.defineData('_negate_maxafter');
		call hnegate.defineDone();
		call missing(_neg_key, _negate_pattern, _negate_maxbefore, _negate_maxafter);
		declare hiter negiter('hnegate');
		call hnegate.add(key: 'X', data: "FREE (?:OF|FROM)", data: 10, data: 10);
		call hnegate.add(key: 'X', data: "FREE (\w+)? (?:OF|FROM)", data: 10, data: 10);
		call hnegate.add(key: 'X', data: "FREE (\w+)? (\w+)? (?:OF|FROM)", data: 10, data: 10);
		call hnegate.add(key: 'X', data: "(?:WITHOUT|NO) EVIDENCE", data: 10, data: 10);
		call hnegate.add(key: 'X', data: "(?:WITHOUT|NO) (\w+)? EVIDENCE", data: 10, data: 10);
		call hnegate.add(key: 'X', data: "(?:WITHOUT|NO) (\w+)? (\w+)? EVIDENCE", data: 10, data: 10);
		call hnegate.add(key: 'X', data: "NOT?", data: 6, data: 3);
		call hnegate.add(key: 'X', data: "(?:HISTORY|PRIOR|PREVIOUS)", data: 5, data: 0);
		call hnegate.add(key: 'X', data: "(?:OLD)", data: 5, data: 5);
		call hnegate.add(key: 'X', data: "(?:NEGATIVE|NA)", data: 0, data: 2);
		call hnegate.add(key: 'X', data: "(?:NEGATIVE) (?:OF|FOR)", data: 5, data: 0);
		call hnegate.add(key: 'X', data: "(?:NEGATIVE) (\w+)? (?:OF|FOR)", data: 5, data: 0);
		call hnegate.add(key: 'X', data: "(?:NON(?!SMALL|HODGKINS))", data: 5, data: 5);
		call hnegate.add(key: 'X', data: "(?:WITHOUT)", data: 4, data: 0);
		call hnegate.add(key: 'X', data: "(?:RESIDUAL|UNCERTAIN)", data: 4, data: 4);
		call hnegate.add(key: 'X', data: "(?:ABSENT|UNINVOLVED|UNREMARKABLE|BENIGN|NONDIAGNOSTIC)", data: 5, data: 5);
		call hnegate.add(key: 'X', data: "CLEAR", data: 0, data: 1);
		call hnegate.add(key: 'X', data: "(?:IS|ARE) CLEAR", data: 0, data: 2);
		call hnegate.add(key: 'X', data: "CLEAR (?:OF|FOR)", data: 5, data: 0);
		call hnegate.add(key: 'X', data: "CLEAR (\w+)? (?:OF|FOR)", data: 5, data: 0);
		call hnegate.add(key: 'X', data: "RULE OUT", data: 5, data: 5);
		call hnegate.add(key: 'X', data: "NOT (?:SUPPORT(?:ED)?|IDENTIFIED)", data: 7, data: 7);
		call hnegate.add(key: 'X', data: "IMAGING", data: 0, data: 3);
		call hnegate.add(key: 'X', data: "UNLIKELY", data: 5, data: 5);
		call hnegate.add(key: 'X', data: "LESS LIKELY", data: 5, data: 5);
		call hnegate.add(key: 'X', data: "FAILS? TO", data: 5, data: 0);
		call hnegate.add(key: 'X', data: "LINE", data: 0, data: 1);
		call hnegate.add(key: 'X', data: "STATUS POST", data: 5, data: 5);
		call hnegate.add(key: 'X', data: "STATUS (\w+)? POST", data: 5, data: 5);
		call hnegate.add(key: 'X', data: "STATUS (\w+)? (\w+)? POST", data: 5, data: 5);
		call hnegate.add(key: 'X', data: "TO EXCLUDE", data: 5, data: 0);
		call hnegate.add(key: 'X', data: "EVALUATE", data: 5, data: 0);
		call hnegate.add(key: 'X', data: "CLINICAL INDICATION", data: 5, data: 0);
		call hnegate.add(key: 'X', data: "LOW (?:LIKELIHOOD|PROBABILITY)", data: 5, data: 0);
		call hnegate.add(key: 'X', data: "DONT SEE", data: 6, data: 0);
		call hnegate.add(key: 'X', data: "DO NOT SEE", data: 6, data: 0);
		call hnegate.add(key: 'X', data: "NO (\w+)? (?:SUGGEST(?:ING)?)", data: 6, data: 0);
		call hnegate.add(key: 'X', data: "NO (\w+)? (\w+)? (?:SUGGEST(?:ING)?)", data: 6, data: 0);
		call hnegate.add(key: 'X', data: "NO (\w+)? (\w+)? (\w+)? (?:SUGGEST(?:ING)?)", data: 6, data: 0);
		call hnegate.add(key: 'X', data: "NO (\w+)? (\w+)? (\w+)? (\w+)? (?:SUGGEST(?:ING)?)", data: 6, data: 0);
		call hnegate.add(key: 'X', data: "NO (\w+)? (\w+)? (\w+)? (\w+)? (\w+)? (?:SUGGEST(?:ING)?)", data: 6, data: 0);
	end;

	/*MATCH TNM STAGING & EXTRACT*/
	_tnmpattern=prxparse("/(?i)(?:AJCC|Y?P?TP?NC?M|STAG(?:E|ING)).{0,30}?\sY?P?T([0-9][A-Z]?|C?IS|X).{0,10}?P?N([0-9]|X).{0,10}?C?M([0-9]|X)/");
	if (prxmatch(_tnmpattern,&var)>0) then do;
		nlp_match_ajcc=1;
      	call prxposn(_tnmpattern, 1, _tnmpos, _tnmlen);
      	_tnmt=substr(&var, _tnmpos, _tnmlen);
      	call prxposn(_tnmpattern, 2, _tnmpos, _tnmlen);
      	_tnmn=substr(&var, _tnmpos, _tnmlen);
      	call prxposn(_tnmpattern, 3, _tnmpos, _tnmlen);
      	_tnmm=substr(&var, _tnmpos, _tnmlen);
		nlp_ajcc="T" || strip(_tnmt) || " N" || strip(_tnmn) || " M" || strip(_tnmm);
		nlp_ajcc_t=strip(_tnmt);
		nlp_ajcc_n=strip(_tnmn);
		nlp_ajcc_m=strip(_tnmm);
	end; else nlp_match_ajcc=0;
	call prxfree(_tnmpattern);

%if (%upcase(&cancer)=PCA) %then %do;
	/*MATCH GLEASON SCORE & EXTRACT*/
	nlp_match_gleason=0;
	regex_gleason1=prxparse("/(?:GLE?A?S?O?N?\'?S?|GLS|GL.{1,10}?(?:SCOR|PAT|GRA?D?E?)|SCORE|G[GS]).{0,10}?([1-5])\s?[\+\,\/\-\~]+\s?([1-5])\s?[\+\,\/\-\=\~]+\s?(10|[2-9])/i");
	regex_gleason2=prxparse("/(?:GLE?A?S?O?N?\'?S?|GLS|GL.{1,10}?(?:SCOR|PAT|GRA?D?E?)|SCORE|G[GS]).{0,10}?([1-5])\s?[\+\,\/\-\=\~]+\s?([1-5])\s?[\+\,\/\-\=\~]+\s?(10|[6-9])/i");
	regex_gleason3=prxparse("/(?:GLE?A?S?O?N?\'?S?|GLS|GL.{1,10}?(?:SCOR|PAT|GRA?D?E?)|SCORE|G[GS]).{0,10}?([1-5])\s?[\+\,\/\-\~]+\s?([1-5])/i");
	regex_gleason4=prxparse("/(?:GLE?A?S?O?N?\'?S?|GLS|GL.{1,10}?(?:SCOR|PAT|GRA?D?E?)|SCORE|G[GS]).{0,10}?\=\s?(10|[2-9])/i");
	regex_gleason5=prxparse("/(?:GLE?A?S?O?N?\'?S?|GLS|GL.{1,10}?(?:SCOR|PAT|GRA?D?E?)|SCORE|G[GS]).{0,5}?(10|[2-9])(?!\+)/i");
	regex_gleason6=prxparse("/(?:GLE?A?S?O?N?\'?S?|GLS|GL.{1,10}?(?:SCOR|PAT|GRA?D?E?)|SCORE|G[GS]).{0,10}?SCORE OF (VIII|VII|VI|IX|XII|III|II|IV|I|V|X)/");
	if (prxmatch(regex_gleason1,&var)) then do;
		nlp_match_gleason=1;
		call prxposn(regex_gleason1, 1, _start, _len);
		nlp_gleason_one=input(substr(&var,_start,_len),8.0);
		call prxposn(regex_gleason1, 2, _start, _len);
		nlp_gleason_two=input(substr(&var,_start,_len),8.0);
		nlp_gleason=nlp_gleason_one+nlp_gleason_two;
	end; else if (prxmatch(regex_gleason2,&var)) then do;
		nlp_match_gleason=1;
		call prxposn(regex_gleason2, 1, _start, _len);
		nlp_gleason_one=input(substr(&var,_start,_len),8.0);
		call prxposn(regex_gleason2, 2, _start, _len);
		nlp_gleason_two=input(substr(&var,_start,_len),8.0);
		nlp_gleason=nlp_gleason_one+nlp_gleason_two;
	end; else if (prxmatch(regex_gleason3,&var)) then do;
		nlp_match_gleason=1;
		call prxposn(regex_gleason3, 1, _start, _len);
		nlp_gleason_one=input(substr(&var,_start,_len),8.0);
		call prxposn(regex_gleason3, 2, _start, _len);
		nlp_gleason_two=input(substr(&var,_start,_len),8.0);
		nlp_gleason=nlp_gleason_one+nlp_gleason_two;
	end; else if (prxmatch(regex_gleason4,&var)) then do;
		nlp_match_gleason=1;
		call prxposn(regex_gleason4, 1, _start, _len);
		nlp_gleason=input(substr(&var,_start,_len),8.0);
	end; else if (prxmatch(regex_gleason5,&var)) then do;
		nlp_match_gleason=1;
		call prxposn(regex_gleason5, 1, _start, _len);
		nlp_gleason=input(substr(&var,_start,_len),8.0);
	end; else if (prxmatch(regex_gleason6,&var)) then do;
		nlp_match_gleason=1;
		call prxposn(regex_gleason6, 1, _start, _len);
		_nlp_gleason_txt=substr(&var,_start,_len);
		if (upcase(strip(_nlp_gleason_txt))="I") then nlp_gleason=1;
		else if (upcase(strip(_nlp_gleason_txt))="II") then nlp_gleason=2;
		else if (upcase(strip(_nlp_gleason_txt))="III") then nlp_gleason=3;
		else if (upcase(strip(_nlp_gleason_txt))="IV") then nlp_gleason=4;
		else if (upcase(strip(_nlp_gleason_txt))="V") then nlp_gleason=5;
		else if (upcase(strip(_nlp_gleason_txt))="VI") then nlp_gleason=6;
		else if (upcase(strip(_nlp_gleason_txt))="VII") then nlp_gleason=7;
		else if (upcase(strip(_nlp_gleason_txt)) in ('IIX','VIII')) then nlp_gleason=8;
		else if (upcase(strip(_nlp_gleason_txt))="IX") then nlp_gleason=9;
		else if (upcase(strip(_nlp_gleason_txt))="X") then nlp_gleason=10;
	end;
	call prxfree(regex_gleason1);
	call prxfree(regex_gleason2);
	call prxfree(regex_gleason3);
	call prxfree(regex_gleason4);
	call prxfree(regex_gleason5);
	call prxfree(regex_gleason6);
%end;

	/*MAIN NLP ALGORITHM*/
	regex_start=1;
	regex_stop=lengthn(&var);
	if (regex_stop=0) then return;
	call prxnext(regex_line, regex_start, regex_stop, &var, regex_position, regex_length);
	do while (regex_position>0 & regex_length>0);
		call h.clear();
		if (regex_position>1) then do;
			regex_position=regex_position+4;
			regex_length=regex_length-4;
		end;
		line=substr(&var, regex_position, regex_length);
		_newline='';

		_lineforarray=prxchange("s/\'//i",-1,line);
		_lineforarray=prxchange('s/[[:punct:]]/ /i',-1,_lineforarray);
		_lineforarray=prxchange('s/\s{2,}/ /i',-1,_lineforarray);
		_linearraywords=countw(strip(_lineforarray)," ");
		do _linewordcount=1 to dim(_line_array);
			_line_array{_linewordcount}='';
		end;
		do _linewordcount=1 to _linearraywords;
			_line_array{_linewordcount}=scan(strip(_lineforarray), _linewordcount, " ");
			if (prxmatch('/(?<=\b)(?:A|ALSO|AND|AT|BY|FOR|FROM|HAS|IN|IS|IT|LIKE|OF|OR|OUT|THAT|THE|THIS|TO|WITH)(?=\b)/i', strip(_line_array{_linewordcount}))) then _skip_array{_linewordcount}=1;
			else _skip_array{_linewordcount}=0;
		end;
		
		/*LOOP THROUGH EACH CONCEPT FROM THE NLP DICTIONARY*/
		_rc=iter.first();
		do while(_rc=0);
			regex_match=0;
			array txt_pos{6} txt_pos1-txt_pos6;
			*if (snm_code="M8124X") then putlog snm_code= line=;

			/*LOOP THROUGH EACH WORD FOR THE CURRENT CONCEPT AND SEARCH FOR NON-NEGATED MATCHES*/
			if (lengthn(txt_neg1)) then do;
				call hnegate.add(key: 'T', data: strip(txt_neg1), data: 5, data: 5);
			end;
			do _tcount=1 to dim(txt_pos);
				if (lengthn(strip(txt_pos{_tcount}))=0) then leave; /*END LOOP IF NO ADDITIONAL CONCEPT WORDS EXIST*/
				if (h.check(key: strip(txt_pos{_tcount}))=0) then do;
					if (h.find(key: strip(txt_pos{_tcount}))=0) then do;
						if (missing(_hpos)=1) then do;
							if (strip(txt_pos{_tcount})="CELL") then do;
								continue;
							end; else do;
								regex_match=0;
								leave;
							end;
						end; else do;
							regex_match=1;
							continue;
						end;
					end;
				end;
				_anymatch=0;
				do _linewordcount=1 to _linearraywords;
					if (_skip_array{_linewordcount}) then continue;
					length _matchword $50;
					_matchword=strip(txt_pos{_tcount});
					if (lengthn(_matchword)>=4) then do;
						if (upcase(substr(_matchword,lengthn(_matchword),1))="S") then _matchword=strip(_matchword) || "(?:ES)?";
						else if (upcase(substr(_matchword,lengthn(_matchword),1))="Y") then _matchword=substr(_matchword,1,lengthn(_matchword)-1) || "(:?Y|IES)";
						else _matchword=strip(_matchword) || "S?";
					end;
					if (prxmatch("/(?<=\b)(" || strip(_matchword) || ")(?=\b)" || "/i", strip(_line_array{_linewordcount}))) then do;
						_negated=0;
						_nc=negiter.first();
						do while(_nc=0);
							_negate_pattern_words=countw(strip(_negate_pattern)," ");
							_negate_before_andor=0;
							_negate_after_andor=0;
							if ((_linewordcount-_negate_maxbefore)<1) then _andcheckmin=1;
							else _andcheckmin=_linewordcount-_negate_maxbefore;
							if ((_linewordcount+_negate_maxafter)>_linearraywords) then _andcheckmax=_linearraywords;
							else _andcheckmax=_linewordcount+_negate_maxafter;
							do _andcount=_andcheckmin to _andcheckmax;
								if (strip(_line_array{_andcount}) in ('AND','OR','NOR','WITH')) then do;
									if (_andcount<=_linewordcount) then _negate_before_andor=_negate_before_andor+1;
									else _negate_after_andor=_negate_after_andor+1;
								end;
							end;
							do _negatecount=1 to (_linearraywords-_negate_pattern_words+1);
								if ((_linewordcount-_negatecount)>(_negate_maxbefore+((_negate_maxbefore+3)*_negate_before_andor))) then do;
									continue;
								end;
								if ((_negatecount-_linewordcount)>(_negate_maxafter+((_negate_maxafter+3)*_negate_after_andor))) then do;
									leave;
								end;
								do _negateinner=1 to _negate_pattern_words;
									_negate_word_curr=scan(_negate_pattern, _negateinner, " ");
									if ((strip(_negate_word_curr)="(\w+)?") | (prxmatch("/(?<=\b)" || strip(_negate_word_curr) || "(?=\b)/i",_line_array{_negatecount+_negateinner-1}))) then do;
										_negated=1;
									end; else do;
										_negated=0;
										leave;
									end;
								end;
								if (_negated=1) then leave;
							end;
							if (_negated=1) then leave;
							_nc=negiter.next();
						end;
						if (_negated~=1) then do;
							_anymatch=1;
							regex_match=1;
							call h.add(key: strip(txt_pos{_tcount}), data: _linewordcount);
						end;
					end;
				end;
				if (_anymatch~=1) then do;
					call h.add(key: strip(txt_pos{_tcount}), data: .);
					if (strip(_matchword)~="CELLS?") then do;
						regex_match=0;
						leave;
					end;
				end;
			end;
			if (lengthn(txt_neg1)) then do;
				call negiter.last();
				call hnegate.remove(key: 'T');
			end;
			/*END LOOPING THROUGH WORDS OF CURRENT CONCEPT*/
			
			if (regex_match=1) then do;
				do _inner_tcount=1 to dim(txt_pos);
					if (lengthn(txt_pos{_inner_tcount})=0) then leave;
					if (h.find(key: strip(txt_pos{_inner_tcount}))=0) then do;
						if (missing(_hpos)=1) then do;
							if (strip(txt_pos{_inner_tcount})~="CELL") then do;
								regex_match=0;
								leave;
							end;
						end;
					end; else regex_match=0;
				end;

				do _cleari=1 to 10;
					do _clearj=1 to 10;
						_word_pos{_cleari,_clearj}=.;
					end;
				end;					
				if (regex_match=1) then do;
					do _inner_tcount=1 to dim(txt_pos);
						if (lengthn(txt_pos{_inner_tcount})=0) then leave;
						_poscount=1;
						if (h.find(key: strip(txt_pos{_inner_tcount}))=0) then do;
							if ((missing(_hpos)) & (strip(txt_pos{_inner_tcount})="CELL")) then do;
								continue;
							end;
							_word_pos{_inner_tcount,_poscount}=_hpos;
							h.has_next(result: _hnext);
							do while(_hnext~=0);
								call h.find_next();
								_poscount=_poscount+1;
								_word_pos{_inner_tcount,_poscount}=_hpos;
								h.has_next(result: _hnext);
							end;
						end;
					end;

					array _matchpos{10} 8;
					do _tmpcount=1 to dim(_word_pos,2);
						do _tmpclear=1 to dim(_matchpos);
							_matchpos{_tmpclear}=.;
						end;						
						if (missing(_word_pos{1,_tmpcount})) then leave;
						_min_pos=.;
						_max_pos=.;
						_skipcell=0;
						_tmpmatch=1;
						do _inner_tcount=1 to dim(txt_pos);
							if (lengthn(txt_pos{_inner_tcount})=0) then leave;
							_tmpmatch=0;
							do _tmpinner=1 to dim(_word_pos,2);
								if (missing(_word_pos{_inner_tcount,_tmpinner})) then do;
									if (strip(txt_pos{_inner_tcount})="CELL") then do;
										_tmpmatch=1;
										_skipcell=1;
										continue;
									end; else do;
										leave;
									end;
								end;
								if (missing(_min_pos)) then _min_pos=_word_pos{_inner_tcount,_tmpinner};
								if (missing(_max_pos)) then _max_pos=_word_pos{_inner_tcount,_tmpinner};
								if ((_word_pos{_inner_tcount,_tmpinner}<=(_max_pos+10)) & (_word_pos{_inner_tcount,_tmpinner}>=(_min_pos-10))) then do;
									_tmpmatch=1;
									_matchpos{_inner_tcount-_skipcell}=_word_pos{_inner_tcount,_tmpinner};
									if (_word_pos{_inner_tcount,_tmpinner}>_max_pos) then _max_pos=_word_pos{_inner_tcount,_tmpinner};
									if (_word_pos{_inner_tcount,_tmpinner}<_min_pos) then _min_pos=_word_pos{_inner_tcount,_tmpinner};
									leave;
								end;
							end;
							if (_tmpmatch=0) then do;
								leave;
							end;
						end;

						if (_tmpmatch=1) then do;
							length _subtype _subclass $1 _probability $5;
							_subclass="X";
							_probability="X";
							if (upcase(strip(snm_type))='MORPHOLOGY') then do;
								_subtype="M";
								_subclass="3";
								_probability="1.0";
								do _subclasscount=1 to _linearraywords;
									if ((_subclasscount<=(_max_pos+5)) & (_subclasscount>=(_min_pos-5))) then do;
										if (_subclass~="1" & prxmatch("/(?i)METASTA[ST]/",_line_array{_subclasscount})) then _subclass="6";
										else if (_subclass~="1" & prxmatch("/(?i)(?:UNKNOWN|UNCERTAIN)/",_line_array{_subclasscount})) then _subclass="9";
										else if (_subclass="3" & prxmatch("/(?i)(?:(?:IN.?)?SITU|NON.?INVASIVE|NON.?INFILTRATING|INTRAEPITHELIAL|DCIS)/",_line_array{_subclasscount})) then _subclass="2";
										if (prxmatch("/(?i)(?:BORDERLINE|SUSPICIOUS|SUGGEST(?:S|IVE)?|POSSIB(?:LE|ILITY))/",_line_array{_subclasscount})) then _probability="0.5";
										else if ((_linearraywords-_subclasscount)>=1) then do;
											if (prxmatch("/(?i)(?:COULD|MIGHT|MAY)/",_line_array{_subclasscount}) & prxmatch("/(?i)(?:INDICATE|REPRESENT|BE)/",_line_array{_subclasscount+1})) then _probability="0.5";
											else if (prxmatch("/(?i)(?:UNCLEAR)/",_line_array{_subclasscount}) & prxmatch("/(?i)(?:WHETHER)/",_line_array{_subclasscount+1})) then _probability="0.5";
										end;
										else if ((_linearraywords-_subclasscount)>=2) then do;
											if (prxmatch("/(?i)(?:CANNOT|CANT)/",_line_array{_subclasscount}) & prxmatch("/(?i)(?:BE)/",_line_array{_subclasscount+1}) & prxmatch("/(?i)(?:EXCLUDED)/",_line_array{_subclasscount+2})) then _probability="0.5";
										end;
									end;
								end;
								if (_subclass="3" & substr(snm_code,lengthn(snm_code),1)~="3") then do;
									if (substr(snm_code,lengthn(snm_code),1)="1") then _subclass="1";
									else if (substr(snm_code,lengthn(snm_code),1)="2") then _subclass="2";
									else if (substr(snm_code,lengthn(snm_code),1)="6") then _subclass="6";
									else if (substr(snm_code,lengthn(snm_code),1)="9") then _subclass="9";
								end;
								if (upcase(substr(snm_code,lengthn(snm_code),1))="X") then snm_code=substr(snm_code,1,lengthn(snm_code)-1) || strip(_subclass);
							end; else if (upcase(strip(snm_type))="PROCEDURE") then _subtype="P";
							else if (upcase(strip(snm_type))="TOPOLOGY") then _subtype="T";
							else _subtype="X";

							if (lengthn(nlp_subtypes)=0) then nlp_subtypes=strip(_subtype);
							else nlp_subtypes=strip(nlp_subtypes) || "|" || strip(_subtype);
							if (lengthn(nlp_subclasses)=0) then nlp_subclasses=strip(_subclass);
							else nlp_subclasses=strip(nlp_subclasses) || "|" || strip(_subclass);
							if (lengthn(nlp_pvals)=0) then nlp_pvals=strip(_probability);
							else nlp_pvals=strip(nlp_pvals) || "|" || strip(_probability);
							if (lengthn(nlp_codes_all)=0) then nlp_codes_all=strip(snm_code);
							else nlp_codes_all=strip(nlp_codes_all) || "|" || strip(snm_code);
							if (prxmatch("/\b" || strip(snm_code) || "\b/i", nlp_codes)=0) then do;
								if (lengthn(nlp_codes)=0) then nlp_codes=strip(snm_code);
								else nlp_codes=strip(nlp_codes) || "|" || strip(snm_code);
							end;

							do _tmpclear=1 to dim(_matchpos);
								if (missing(_matchpos{_tmpclear})) then leave;
								_tmpkey=strip(put(_tcount-1,8.0)) || strip(put(_linearraywords-_matchpos{_tmpclear},8.0)) || strip(snm_code);
								call hmatch.add(key: strip(_tmpkey), data: _matchpos{_tmpclear}, data: strip(snm_code), data: strip(_subtype), data: strip(_subclass), data: strip(_probability));
							end;
							leave;
						end;
					end;
				end;
			end;

			_rc=iter.next();
		end;

		call htaken.clear();
		_mc=miter.first();
		do while(_mc=0);
			if (htaken.find(key: strip(put(_mpos,8.0)))) then do;
				if (strip(_mclass)="X") then _tagtext=" SNM=" || strip(_mval) || " TYPE=" || strip(_msub);
				else _tagtext=" SNM=" || strip(_mval) || " TYPE=" || strip(_msub) || " CLASS=" || strip(_mclass) || " PVAL=" || strip(_mpval);
				_line_array{_mpos}="<NLP" || trim(_tagtext) || ">" || trim(_line_array{_mpos}) || "</NLP" || trim(_tagtext) || ">";
				call htaken.add(key: strip(put(_mpos,8.0)), data: 1);
			end;
			_mc=miter.next();
		end;
		call hmatch.clear();

		do _newlinecount=1 to _linearraywords;
			if (missing(_line_array{_newlinecount})) then leave;
			if (lengthn(_newline)=0) then _newline=_line_array{_newlinecount};
			else _newline=strip(_newline) || " " || _line_array{_newlinecount};
		end;

		if (lengthn(coded)=0) then coded=strip(_newline) || ".";
		else coded=strip(coded) || "<BR>" || strip(_newline) || ".";

		call prxnext(regex_line, regex_start, regex_stop, &var, regex_position, regex_length);
	end;
	call prxfree(regex_line);

	coded=prxchange("s/\<\/NLP SNM=(\w+) TYPE=[A-Z]( CLASS=[0-9])?( PVAL=[0-9]\.[0-9]*)?\>((?:\s|\,|IN|OF|THE|FOR|WITH)*)\<NLP SNM=\1 TYPE=[A-Z]\2?\3?\>/$4/",-1,coded);
	coded=prxchange("s/(\<\/NLP SNM=\w+) TYPE=[A-Z]( CLASS=[0-9])?( PVAL=[0-9]\.[0-9]*)?\>/$1>/",-1,coded);

	_refc=iter.first();
	do while(_refc=0);
		if (upcase(snm_class) in ('BASAL','BENIGN')) then do;
			length _codetorep $10;
			_codetorep=snm_code;
			if (upcase(substr(_codetorep,lengthn(_codetorep),1))="X") then _codetorep=substr(snm_code,1,lengthn(snm_code)-1) || "[0-9]";
			coded=prxchange("s/<NLP SNM=" || strip(_codetorep) || " TYPE=[A-Z](?: CLASS=[0-9])?( PVAL=[0-9]\.[0-9]*)?>//i",-1,coded);
			coded=prxchange("s/<\/NLP SNM=" || strip(_codetorep) || ">//i",-1,coded);
		end;
		_refc=iter.next();
	end;

	log_pctdone=(_N_/log_count)*100;
	if (log_pctdone>=(log_lastpct+1) & log_pctdone<100) then do;
		log_lastpct=round((_N_/log_count)*100,1.0);
		log_timepass=datetime()-log_start;
		log_n=put(_N_,8.0);
		if (log_timepass<=60) then log_text=strip(put(_N_,COMMA8.0)) || " of " || strip(put(log_count,COMMA8.0)) || " (" || strip(put(log_lastpct,8.0)) || "%)" || " in " || strip(put(log_timepass,4.1)) || " seconds.";
		else log_text=strip(put(_N_,COMMA8.0)) || " of " || strip(put(log_count,COMMA8.0)) || " (" || strip(put(log_lastpct,8.0)) || "%)" || " in " || strip(put(log_timepass/60,5.1)) || " minutes.";
		put "NOTE: Finished coding observation #" log_text;
	end;
run;

data replacecodes(keep=snm_code rename=(snm_code=_replace_code));
	set lookup_scent;
	if (txt_pos1="(?:CARCINOMA|CANCER|TUMOU?R|METASTAS[EI]S)" & snm_code~="M8010X");
run;

proc sql noprint;
	select max(countw(nlp_codes_all,"|")) into :codecount from temp;
quit;

data temp(compress=yes drop=_:);
	set temp;
	if (_N_=1) then do;
		length _replace_code $6;
		declare hash hreference(dataset:"replacecodes");
		call hreference.defineKey('_replace_code');
		call hreference.defineData('_replace_code');
		call hreference.defineDone();
		call missing(_replace_code);
		declare hiter iter('hreference');
	end;
	array _nlp_code{&codecount} $10;
	array _nlp_subtype{&codecount} $1;
	array _nlp_subclass{&codecount} $1;
	array _nlp_pval{&codecount} $5;
	do _i=1 to countw(nlp_codes_all,"|");
		_nlp_code{_i}=scan(nlp_codes_all,_i,"|");
		_nlp_subtype{_i}=scan(nlp_subtypes,_i,"|");
		_nlp_subclass{_i}=scan(nlp_subclasses,_i,"|");
		_nlp_pval{_i}=scan(nlp_pvals,_i,"|");
	end;
	_rc=iter.first();
	_found=0;
	do while(_rc=0 & _found=0);
		do _i=1 to dim(_nlp_code);
			if (substr(_nlp_code{_i},1,5)=substr(_replace_code,1,5)) then do;
				do _j=1 to dim(_nlp_code);
					if (substr(_nlp_code{_j},1,5)='M8010') then do;
						_nlp_code{_j}='';
						_nlp_subtype{_j}='';
						_nlp_subclass{_j}='';
						_nlp_pval{_j}='';
						do _k=(_j+1) to dim(_nlp_code);
							_nlp_code{_k-1}=_nlp_code{_k};
							_nlp_subtype{_k-1}=_nlp_subtype{_k};
							_nlp_subclass{_k-1}=_nlp_subclass{_k};
							_nlp_pval{_k-1}=_nlp_pval{_k};
							_nlp_code{_k}='';
							_nlp_subtype{_k}='';
							_nlp_subclass{_k}='';
							_nlp_pval{_k}='';
						end;
						if (substr(_nlp_code{_j},1,5)='M8010') then _j=_j-1;
					end;
				end;
				_found=1;
				leave;
			end;
		end;
		_rc=iter.next();
	end;
	nlp_codes_all='';
	nlp_codes='';
	nlp_subtypes='';
	nlp_subclasses='';
	nlp_pvals='';
	do _i=1 to dim(_nlp_code);
		if (lengthn(_nlp_code{_i})=0) then leave;
		if (_i=1) then do;
			nlp_codes_all=_nlp_code{_i};
			nlp_codes=_nlp_code{_i};
			nlp_subtypes=_nlp_subtype{_i};
			nlp_subclasses=_nlp_subclass{_i};
			nlp_pvals=_nlp_pval{_i};
		end; else do;
			nlp_codes_all=strip(nlp_codes_all) || "|" || strip(_nlp_code{_i});
			nlp_subtypes=strip(nlp_subtypes) || "|" || strip(_nlp_subtype{_i});
			nlp_subclasses=strip(nlp_subclasses) || "|" || strip(_nlp_subclass{_i});
			nlp_pvals=strip(nlp_pvals) || "|" || strip(_nlp_pval{_i});
			if (index(nlp_codes,_nlp_code{_i})=0) then nlp_codes=strip(nlp_codes) || "|" || strip(_nlp_code{_i});
		end;
	end;
run;

%if (%length(&defs)>0) %then %do;
	/*import the SNOMED lookup dictionary*/
	libname xls "&defs";
	data snomed_defs(keep=_defs_code _defs_type _defs_class);
		set xls.'SNOMED$'n(rename=(snm_code=_defs_code snm_type=_defs_type snm_class=_defs_class));
	run;
	libname xls clear;
	data temp(compress=yes drop=_:);
		set temp;
		length _currcode $10 _currpval $10 nlp_result $50 nlp_classes $1000 _code_class _nlp_result 8;
		if (_N_=1) then do;
			length _defs_code _defs_class $20;
			declare hash hdefs(dataset:"snomed_defs");
			call hdefs.defineKey('_defs_code');
			call hdefs.defineData('_defs_code');
			call hdefs.defineData('_defs_class');
			call hdefs.defineDone();
			call missing(_defs_code, _defs_class);
		end;
		nlp_result="Benign";
		_nlp_result=0;
		_code_class=0;
		do _i=1 to countw(nlp_codes_all,"|");
			_currcode=scan(nlp_codes_all,_i,"|");
			_currcodeX=substr(scan(nlp_codes_all,_i,"|"),1,5) || "X";
			_currpval=scan(nlp_pvals,_i,"|");
			if (hdefs.find(key: strip(_currcode))=0) then do;
				if (lengthn(nlp_classes)=0) then nlp_classes=_defs_class;
				else nlp_classes=strip(nlp_classes) || "|" || _defs_class;
				if (prxmatch("/BENIGN/i",_defs_class)) then _code_class=0;
				else if (prxmatch("/BASAL/i",_defs_class) & strip(_currpval)="0.5") then _code_class=1;
				else if (prxmatch("/BASAL/i",_defs_class) & strip(_currpval)="1.0") then _code_class=2;
				else if (prxmatch("/BORDERLINE/i",_defs_class) & strip(_currpval)="0.5") then _code_class=3;
				else if (prxmatch("/BORDERLINE/i",_defs_class) & strip(_currpval)="1.0") then _code_class=4;
				else if (prxmatch("/MALIGNANT/i",_defs_class) & strip(_currpval)="0.5") then _code_class=5;
				else if (prxmatch("/MALIGNANT/i",_defs_class) & strip(_currpval)="1.0") then _code_class=6;		
				if (_code_class in (4,6) & upcase(substr(_currcode,1,5))="M8010" & indexw(strip(upcase(nlp_codes_all)),"P11412","|")) then do;
					_code_class=_code_class-1;
				end;
				if (_code_class>_nlp_result) then _nlp_result=_code_class;
			end; else if (prxmatch("/\A(M[0-9]{4}[12369])\s*\Z/i",_currcode)>0 & hdefs.Find(Key:_currcodeX)=0) then do;
				if (lengthn(nlp_classes)=0) then nlp_classes=_defs_class;
				else nlp_classes=strip(nlp_classes) || "|" || _defs_class;
				if (prxmatch("/BENIGN/i",_defs_class)) then _code_class=0;
				else if (prxmatch("/BASAL/i",_defs_class) & strip(_currpval)="0.5") then _code_class=1;
				else if (prxmatch("/BASAL/i",_defs_class) & strip(_currpval)="1.0") then _code_class=2;
				else if (prxmatch("/BORDERLINE/i",_defs_class) & strip(_currpval)="0.5") then _code_class=3;
				else if (prxmatch("/BORDERLINE/i",_defs_class) & strip(_currpval)="1.0") then _code_class=4;
				else if (prxmatch("/MALIGNANT/i",_defs_class) & strip(_currpval)="0.5") then _code_class=5;
				else if (prxmatch("/MALIGNANT/i",_defs_class) & strip(_currpval)="1.0") then _code_class=6;
				if (_code_class in (4,6) & upcase(substr(_currcode,1,5))="M8010" & indexw(strip(upcase(nlp_codes_all)),"P11412","|")) then do;
					_code_class=_code_class-1;
				end;
				if (_code_class>_nlp_result) then _nlp_result=_code_class;
			end;
		end;
		if (_nlp_result=0) then nlp_result="Benign";
		else if (_nlp_result=1) then nlp_result="Basal (Suspicious)";
		else if (_nlp_result=2) then nlp_result="Basal";
		else if (_nlp_result=3) then nlp_result="Borderline (Suspicious)";
		else if (_nlp_result=4) then nlp_result="Borderline";
		else if (_nlp_result=5) then nlp_result="Malignant (Suspicious)";
		else if (_nlp_result=6) then nlp_result="Malignant";
	run;
%end;

%if (%length(&defs)>0 & %length(&sites)>0) %then %do;
/*IMPORT SITE CATEGORIES AND CREATE RECODES*/
libname xls "&sites";
data alg_sites(drop=site rename=(snm_code=slookup_code site_new=slookup_site));
	set xls.'SAS$'n;
	Site_New=upcase(site);
	if (Site_New in ("BONE MARROW","SPINE","HIP","PELVIS")) then Site_New="BONE";
	if (Site_New in ("CHEST")) then Site_New="BREAST";
	if (Site_New in ("ARMPIT","GROIN","NECK")) then Site_New="LYMPH NODE";
	if ~(Site_New in (("BREAST","BONE","LIVER","LUNG","BRAIN","BLADDER","COLORECTAL","KIDNEY","LUNG","PANCREAS","PROSTATE","SKIN","UTERINE","LYMPH NODE"))) then Site_New="OTHER";
run;
libname xls clear;

/*APPROXIMATE MALIGNANY EXTENT AND SITE*/
%if (%upcase(&cancer)=%str(BCA)) %then %let primarysite=%str(Breast);
%else %if (%upcase(&cancer)=%str(PCA)) %then %let primarysite=%str(Prostate);
data temp(compress=yes drop=_: slookup_: snm_extent: rename=(snm_site_final=snm_site_txt));
	set temp;
	if (_N_=1) then do;
		length slookup_code $10 slookup_site $25;
		declare hash hreference(dataset:"alg_sites");
		call hreference.defineKey('slookup_code');
		call hreference.defineData('slookup_code');
		call hreference.defineData('slookup_site');
		call hreference.defineDone();
		call missing(slookup_code,slookup_site);
		declare hiter miter('hreference');
	end;
	length _currcode _currclass $10 _lastdigit $1 snm_site snm_origin $25 snm_site_final $100;
	do _i=1 to countw(nlp_codes_all,"|");
		_currcode=scan(nlp_codes_all,_i,"|");
		_currclass=scan(nlp_classes,_i,"|");
		_lastdigit='';
		if ~(upcase(_currclass) in ("MALIGNANT","BORDERLINE")) then continue;
		if (prxmatch('/(?:\A|\s)M[0-9XY]{4}([0-9])[1-3](?:\s|\z)/i',_currcode)) then _lastdigit=prxchange("s/(?i)(?:\A|\s)M[0-9XY]{4}([0-9])[1-3](?:\s|\z)/$1/",1,_currcode);
		else if (prxmatch('/(?:\A|\s)M[0-9XY]*?([0-9])(?:\s|\z)/i',_currcode)) then _lastdigit=prxchange("s/(?i)(?:\A|\s)M[0-9XY]*?([0-9])(?:\s|\z)/$1/",1,_currcode);
		if (missing(_lastdigit)=0) then do;
			if (_lastdigit='1' & missing(snm_extent)=1) then snm_extent=1;
			else if (_lastdigit='2' & (missing(snm_extent)=1 | snm_extent=1)) then snm_extent=2;
			else if (_lastdigit in ('3','7','8') & (missing(snm_extent)=1 | snm_extent=1 | snm_extent=2)) then snm_extent=3;
			else if (_lastdigit='9' & (missing(snm_extent)=1 | snm_extent=1 | snm_extent=2 | snm_extent=3)) then snm_extent=9;
			else if (_lastdigit='6' & (missing(snm_extent)=1 | snm_extent=1 | snm_extent=2 | snm_extent=3 | snm_extent=9)) then snm_extent=6;
		end;
	end;

	if (missing(snm_extent)=0 & prxmatch("/(?:BORDERLINE|MALIGNANT|BASAL)/i",nlp_result)) then do;
		if (snm_extent in (6,9)) then do;
			snm_site='UNKNOWN';
			snm_origin='UNKNOWN';
			do _i=1 to countw(nlp_codes_all,"|");
				_currcode=scan(nlp_codes_all,_i,"|");
				if (prxmatch("/(?:\A)T/i",_currcode)=0) then continue;
				if (hreference.find(key: strip(_currcode))=0) then do;
					if (slookup_site="%upcase(&primarysite)") then snm_origin=slookup_site;
					else if (~(slookup_site in ("LYMPH NODE","UNKNOWN")) & snm_site="LYMPH NODE") then snm_origin=slookup_site;
					else if (slookup_site="LYMPH NODE" & ~(snm_site in ("UNKNOWN","LYMPH NODE")) & lengthn(snm_origin)=0) then do;
						snm_origin=snm_site;
						snm_site=slookup_site;
					end; else if (slookup_site="LYMPH NODE") then snm_site=slookup_site;
					else if (snm_site="UNKNOWN") then snm_site=slookup_site;
					else if (snm_site="OTHER" & slookup_site~="LYMPH NODE") then snm_site=slookup_site;
					else if ((snm_site in ("SKIN","LYMPH NODE")) & ~(slookup_site in ("OTHER","LYMPH NODE"))) then snm_site=slookup_site;
				end;
			end;
			if (snm_site in ("UNKNOWN","OTHER")) then do;
				if (prxmatch("/(?:\A|\|)(?:P184347)(?:\||\z)/i",nlp_codes_all)) then snm_site="UTERINE";
				else if (prxmatch("/(?:\A|\|)(?:P103005|P11061)(?:\||\z)/i",nlp_codes_all)) then snm_site="BREAST";
				else if (prxmatch("/(?:\A|\|)(?:P11062)(?:\||\z)/i",nlp_codes_all)) then snm_site="PROSTATE";
				else if (prxmatch("/(?:\A|\|)(?:P11412)(?:\||\z)/i",nlp_codes_all)) then snm_site="SKIN";
			end; else if (snm_site="LYMPH NODE") then do;
				if (prxmatch("/(?:\A|\|)(?:P184347)(?:\||\z)/i",nlp_codes_all)) then snm_origin="UTERINE";
				else if (prxmatch("/(?:\A|\|)(?:P103005|P11061)(?:\||\z)/i",nlp_codes_all)) then snm_origin="BREAST";
				else if (prxmatch("/(?:\A|\|)(?:P11062)(?:\||\z)/i",nlp_codes_all)) then snm_origin="PROSTATE";
				else if (prxmatch("/(?:\A|\|)(?:P11412)(?:\||\z)/i",nlp_codes_all)) then snm_origin="SKIN";
			end;
		end; else do;
			snm_origin='';
			snm_site='UNKNOWN';
			do _i=1 to countw(nlp_codes_all,"|");
				_currcode=scan(nlp_codes_all,_i,"|");
				if (prxmatch("/(?:\A)T/i",_currcode)=0) then continue;
				if (hreference.find(key: strip(_currcode))=0) then do;
					if (slookup_site="%upcase(&primarysite)") then snm_site=slookup_site;
					else if (snm_site="UNKNOWN") then snm_site=slookup_site;
					else if (snm_site="OTHER" & slookup_site~="LYMPH NODE") then snm_site=slookup_site;
					else if ((snm_site in ("SKIN","LYMPH NODE")) & ~(slookup_site in ("OTHER","LYMPH NODE"))) then snm_site=slookup_site;
				end;
			end;
			if (snm_site in ("UNKNOWN","OTHER")) then do;
				if (prxmatch("/(?:\A|\|)(?:P184347)(?:\||\z)/i",nlp_codes_all)) then snm_site="UTERINE";
				else if (prxmatch("/(?:\A|\|)(?:P103005|P11061)(?:\||\z)/i",nlp_codes_all)) then snm_site="BREAST";
				else if (prxmatch("/(?:\A|\|)(?:P11062)(?:\||\z)/i",nlp_codes_all)) then snm_site="PROSTATE";
				else if (prxmatch("/(?:\A|\|)(?:P11412)(?:\||\z)/i",nlp_codes_all)) then snm_site="SKIN";
			end; else if (snm_site="LYMPH NODE") then do;
				if (prxmatch("/(?:\A|\|)(?:P184347)(?:\||\z)/i",nlp_codes_all)) then snm_origin="UTERINE";
				else if (prxmatch("/(?:\A|\|)(?:P103005|P11061)(?:\||\z)/i",nlp_codes_all)) then snm_origin="BREAST";
				else if (prxmatch("/(?:\A|\|)(?:P11062)(?:\||\z)/i",nlp_codes_all)) then snm_origin="PROSTATE";
				else if (prxmatch("/(?:\A|\|)(?:P11412)(?:\||\z)/i",nlp_codes_all)) then snm_origin="SKIN";
			end;
		end;
	end;
	if (lengthn(snm_origin)=0 & lengthn(snm_site)>0) then snm_site_final=strip(snm_site) || " (LOCAL)";
	else if (snm_site="LYMPH NODE") then snm_site_final=strip(snm_origin) || " (REGIONAL)";
	else if (lengthn(snm_origin)>0) then snm_site_final=strip(snm_site) || " (ORIGIN: " || strip(snm_origin) || ")";
	snm_site_final=propcase(snm_site_final);
	%if (%upcase(&primarysite)=%str(BREAST)) %then %do;
		if (snm_site_final="Breast (Local)" & snm_extent=2) then snm_site_final="Breast (DCIS)";
	%end;
	snm_site=propcase(snm_site);
	snm_origin=propcase(snm_origin);

	if (prxmatch("/MALIGNANT/i",nlp_result)) then do;
		if (prxmatch("/&primarysite/i",snm_site) | prxmatch("/&primarysite/i",snm_origin)) then nlp_result=prxchange('s/(MALIGNANT)/Recurrence/i',-1,nlp_result);
		else nlp_result=prxchange('s/(MALIGNANT)/Primary/i',-1,nlp_result);
	end;
run;
proc sql noprint;
	drop table alg_sites;
quit;
%end;

proc sql;
	create table &out(compress=yes) as select * from temp;
quit;

proc sql noprint;
	drop table lookup_scent, inputset, temp, replacecodes, snomed_defs;
quit;

options &opts;
%mend scent_coder;
