/***************************************************************************/
/*Name:          SCENT, Preprocessing                                      */
/*Location:      Undefined                                                 */
/*Author:        Justin A. Strauss                                         */
/*Purpose:                                                                 */
/*Date:                                                                    */
/*Modified:                                                                */
/*License: Licensed under the Apache License, Version 2.0 (the License).  
You may not use this file except in compliance with the License.  
You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0 
  
Unless required by applicable law or agreed to in writing, software distributed under the License 
is distributed on an AS IS BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
See the License for the specific language governing permissions and limitations under the License. 
*/
/***************************************************************************/

%macro scent_preprocessor(data=,var=,out=);

	%macro GetSysOpts;
	%local ls notes source mprint mlogic implmac;
	%let ls=%sysfunc(getoption(ls,keyword));
	%let notes=%sysfunc(getoption(notes,keyword));
	%let source=%sysfunc(getoption(source,keyword));
	%let mprint=%sysfunc(getoption(mprint,keyword));
	%let mlogic=%sysfunc(getoption(mlogic,keyword));
	%let implmac=%sysfunc(getoption(implmac,keyword));
	&ls &notes &source &mprint &mlogic
	%mend GetSysOpts;

%local opts;
%let opts=%GetSysOpts;
options nomprint nomlogic nosource;

%if (%length(&data)>0) %then %let data=%sysfunc(dequote(&data));
%else %return;
%if (%length(&var)>0) %then %let var=%sysfunc(dequote(&var));
%else %return;
%if (%length(&out)>0) %then %let out=%sysfunc(dequote(&out));
%else %let out=&data;
data &out(compress=yes drop=log_:);
	length preprocessed $20000;
	set &data nobs=log_count;
	retain log_lastpct log_start;
	if (_N_=1) then do;
		log_start=datetime();
		log_lastpct=0;
	end;
	preprocessed=&var;
	/*replace special characters*/
	preprocessed=prxchange('s/(?<=[a-z]<BR>)(\s?[A-Z]+)/.$1/', -1, strip(preprocessed));
	preprocessed=prxchange('s/(?<=[A-Za-z0-9]<BR>)(\s?[A-Z][a-z]+)/.$1/', -1, strip(preprocessed));
	preprocessed=prxchange('s/\-\-/. /i', -1, upcase(strip(preprocessed)));
	preprocessed=prxchange('s/([^\.])<BR>\-/$1.<BR>/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/\(\s/(/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/\s\)/)/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/([0-9])\#/$1%/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/([A-Z]|\s)\#/$1./i', -1, strip(preprocessed));
	preprocessed=prxchange('s/([A-Z0-9])\(/$1 /i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(?:<BR>|\A|\s)(?:[0-9]{1,2}|[A-Z])[\:\)\-]\s?/.<BR>/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/\@//i', -1, strip(preprocessed));
	preprocessed=prxchange('s/\(//i', -1, strip(preprocessed));	
	preprocessed=prxchange('s/\)//i', -1, strip(preprocessed));
	/*fix various things*/
	preprocessed=prxchange('s/\s?<BR>(?:[0-9]{1,2}|[A-Z])[\.\:\)\-]\s?/.<BR>/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(?:<BR>|\A)((?:[A-Z0-9]|\s|[\,\+\=\-\#\(\)\/\:]){1,40}?)<BR>((?:[A-Z0-9]|\s|[\,\+\=\-\#\(\)\/\:]){1,40}?)<BR>/<BR>$1.<BR>$2.<BR>/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(?:<BR>|\A)((?:[A-Z0-9]|\s|[\,\+\=\-\#\(\)\/]){1,25}?\:(?:[A-Z0-9]|\s|[\,\+\=\-\#\(\)\/]){1,25}?)<BR>/<BR>$1.<BR>/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(?:<BR>|\A)[A-Z](?:(?:\s|\,|\-|AND|THRU|THROUGH)+[A-Z])+(?:\.|\:|\)|<BR>)//i', -1, strip(preprocessed));
	/*parse out common abbreviations*/
	preprocessed=prxchange('s/(N\/A)/NA/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(NON.?SMALL)/NONSMALL/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(M\.D\.|D\.R\.|F\.D\.A\.|[AP]\.M\.|[CM]\.?M\.)<BR>/$1.<BR>/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(M\.D\.)/MD/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(D\.R\.|DR\.)/DR/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(F\.D\.A\.)/FDA/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(A\.M\.)/AM/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(P\.M\.)/PM/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(CM\.)/CM/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(MM\.)/MM/i', -1, strip(preprocessed));
	/*create line breaks*/
	preprocessed=prxchange('s/(<BR>|\A|\s)[0-9]{1,2}\:[0-9][0-9]\s?(?:AM|PM)/$1[TIME]/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/((?:<BR>|\A)[A-Z0-9])(?:\)|\:)/$1./i', -1, strip(preprocessed));
	preprocessed=prxchange('s/<BR>/ /', -1, strip(preprocessed));
	preprocessed=prxchange('s/((?:<BR>|\A).*?)\:(.*?)(?=<BR>)/$1:$2./i', -1, strip(preprocessed));
	preprocessed=prxchange('s/((?:<BR>|\A).*?\:.*?)<BR>/$1. /i', -1, strip(preprocessed));
	preprocessed=prxchange('s/((?:[A-Z0-9]|[[:punct:]]))<BR>((?:[A-Z0-9\-\+\,]{2,}|A))/$1 $2/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/([A-Z])<BR>([0-9]\.?[0-9\s])/$1 $2/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/([A-Z0-9])<BR>([B-Z0-9][^\s])/$1. $2/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/([A-Z0-9])<BR>([B-Z0-9][\s])/$1 $2/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/((?:[A-Z]|[[:punct:]]))<BR>([B-Z0-9]\.)/$1.<BR>$2/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/<BR>//i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(.[\.])(?:\s|\-){0,3}([A-Z])/$1<BR>$2/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/((?:[A-Z]|[[:punct:]])[\.])\s?((?:[A-Z0-9\(]|[[:punct:]]))/$1<BR>$2/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/\:\s?([0-9]\.)([^0-9])/:<BR>$1$2/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/((?:[0-9]|[[:punct:]])\.)\s?([0-9]\.)/$1<BR>$2/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/\:([A-Z0-9])/: $1/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/\,([A-Z0-9])/, $1/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(?:<BR>|\A)\(?[A-Z0-9]\)\s?//i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(?:<BR>|\A)(?:.[[:punct:]]?){1,4}<BR>/<BR>/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(?:<BR>|\A).{1,4}\.?\s?<BR>/<BR>/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/\s{2,}/ /i', -1, strip(preprocessed));
	preprocessed=prxchange('s/\s\././i', -1, strip(preprocessed));
	preprocessed=prxchange('s/\.{2,}/./i', -1, strip(preprocessed));
	preprocessed=prxchange('s/\:\./:/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(<BR>|\A|\s)[0-9]{1,2}(?:\:|\-|\))(<BR>|\s)/<BR>/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(?:<BR>|\A)(?:[\.\,\(\)\-]|\s)+/<BR>/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(?:<BR>|\A).{0,4}<BR>/<BR>/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(?:\s)AND\s?[[:punct:]]?<BR>/ AND /i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(?:<BR>)AND\s/ AND /i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(?:<BR>){2,}/<BR>/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/\A(?:<BR>)+//i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(?:<BR>)*\z//i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(?:<BR>)?FDA DISCLAIMER THIS IMMUNOLOGIC TEST WAS DEVELOPED AND ITS PERFORMANCE CHARACTERISTICS DETERMINED BY THE SCPMG IMMUNOHISTOCHEMISTRY LABORATORY.//i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(?:<BR>)?UNLESS INDICATED OTHERWISE, IT HAS NOT BEEN CLEARED OR APPROVED BY THE USFDA, ALTHOUGH SUCH APPROVAL IS GENERALLY NOT REQUIRED FOR AN ANALYTE SPECIFIC REAGENT OF THIS TYPE.//i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(<BR>|\A|\s)DR\s(?:[A-Z]|\-)+/$1[MD]/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(<BR>|\A|\s)(?:(?:[A-Z]|\-)+\s?){1,2}\,?\sMD/$1[MD]/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(<BR>|\A|\s)[0-9]{1,2}[\/\-][0-9]{1,2}[\/\-][0-9]{2,4}/$1[DATE]/i', -1, strip(preprocessed));
	preprocessed=prxchange('s/(<BR>|\A|\s)(?:JAN[A-Z]*|FEB[A-Z]*|MAR[A-Z]*|APR[A-Z]*|MAY|JUNE?|JULY?|AUG[A-Z]*|SEP[A-Z]*|OCT[A-Z]*|NOV[A-Z]*|DEC[A-Z]*)\s?[0-9]{1,2}\,?\s?[0-9]{4}/$1[DATE]/i', -1, strip(preprocessed));

	log_pctdone=(_N_/log_count)*100;
	if (log_pctdone>=(log_lastpct+5) & log_pctdone<100) then do;
		log_lastpct=round((_N_/log_count)*100,1.0);
		log_timepass=datetime()-log_start;
		log_n=put(_N_,8.0);
		if (log_timepass<=60) then log_text=strip(put(_N_,COMMA8.0)) || " of " || strip(put(log_count,COMMA8.0)) || " (" || strip(put(log_lastpct,8.0)) || "%)" || " in " || strip(put(log_timepass,4.1)) || " seconds.";
		else log_text=strip(put(_N_,COMMA8.0)) || " of " || strip(put(log_count,COMMA8.0)) || " (" || strip(put(log_lastpct,8.0)) || "%)" || " in " || strip(put(log_timepass/60,5.1)) || " minutes.";
		put "NOTE: Finished coding observation #" log_text;
	end;
run;

options &opts;
%mend scent_preprocessor;
